import os
import shutil
from collections import OrderedDict
from yggdrasil import platform, tools
from yggdrasil.drivers.CompiledModelDriver import (
    CompiledModelDriver, CompilerBase, ArchiverBase)
from yggdrasil.languages import get_language_dir


_default_internal_libtype = 'object'
# if platform._is_win:  # pragma: windows
#     _default_internal_libtype = 'static'


class CCompilerBase(CompilerBase):
    r"""Base class for C compilers."""
    languages = ['c']
    default_executable_env = 'CC'
    # TODO: Additional flags environment variables?
    default_flags_env = 'CFLAGS'
    default_flags = ['-g', '-Wall']
    # GCC & CLANG have similar call patterns
    linker_attributes = {'default_flags_env': 'LDFLAGS',
                         'search_path_env': ['LIBRARY_PATH', 'LD_LIBRARY_PATH']}
    search_path_env = ['C_INCLUDE_PATH']
    search_path_flags = ['-E', '-v', '-xc', '/dev/null']
    search_regex_begin = '#include "..." search starts here:'
    search_regex_end = 'End of search list.'
    search_regex = [r'(?:#include <...> search starts here:)|'
                    r'(?: ([^\n]+?)(?: \(framework directory\))?)\n']

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        if platform._is_mac:
            cls.linker_attributes = dict(cls.linker_attributes,
                                         search_path_flags=['-Xlinker', '-v'],
                                         search_regex=[r'\t([^\t\n]+)\n'],
                                         search_regex_begin='Library search paths:')
        elif platform._is_linux:
            cls.linker_attributes = dict(cls.linker_attributes,
                                         search_path_flags=['-Xlinker', '--verbose'],
                                         search_regex=[r'SEARCH_DIR\("=([^"]+)"\);'])
        CompilerBase.before_registration(cls)

    @classmethod
    def call(cls, args, **kwargs):
        r"""Call the compiler with the provided arguments. For |yggdrasil| C
        models will always be linked using the C++ linker since some parts of
        the interface library are written in C++."""
        if not kwargs.get('dont_link', False):
            kwargs.setdefault('linker_language', 'c++')
        return super(CCompilerBase, cls).call(args, **kwargs)
    

class GCCCompiler(CCompilerBase):
    r"""Interface class for gcc compiler/linker."""
    toolname = 'gcc'
    platforms = ['MacOS', 'Linux', 'Windows']
    default_archiver = 'ar'


class ClangCompiler(CCompilerBase):
    r"""clang compiler on Apple Mac OS."""
    toolname = 'clang'
    platforms = ['MacOS']
    default_archiver = 'libtool'


class MSVCCompiler(CCompilerBase):
    r"""Microsoft Visual Studio C Compiler."""
    toolname = 'cl'
    languages = ['c', 'c++']
    platforms = ['Windows']
    default_flags_env = ['CFLAGS', 'CXXFLAGS']
    # TODO: Currently everything compiled as C++ on windows to allow use
    # of complex types. Use '/TC' instead of '/TP' for strictly C
    default_flags = ['/W4',      # Display all errors
                     '/Zi',      # Symbolic debug in .pdb (implies debug)
                     # '/MTd',     # Use LIBCMTD.lib to create multithreaded .exe
                     # '/Z7',      # Symbolic debug in .obj (implies debug)
                     "/EHsc",    # Catch C++ exceptions only (C don't throw C++)
                     '/TP',      # Treat all files as C++
                     "/nologo",  # Suppress startup banner
                     # Don't show errors from using scanf, strcpy, etc.
                     "-D_CRT_SECURE_NO_WARNINGS"]
    output_key = '/Fo%s'
    output_first = True
    default_linker = 'LINK'
    default_archiver = 'LIB'
    linker_switch = '/link'
    search_path_env = 'INCLUDE'
    search_path_flags = None
    version_flags = []
    product_exts = ['.dir', '.ilk', '.pdb', '.sln', '.vcxproj', '.vcxproj.filters']
    combine_with_linker = True  # Must be explicit; linker is separate .exe
    linker_attributes = dict(GCCCompiler.linker_attributes,
                             default_executable=None,
                             default_flags_env=None,
                             output_key='/OUT:%s',
                             output_first=True,
                             output_first_library=False,
                             flag_options=OrderedDict(
                                 [('library_libs', ''),
                                  ('library_dirs', '/LIBPATH:%s')]),
                             shared_library_flag='/DLL',
                             search_path_env='LIB',
                             search_path_flags=None)
    
    @classmethod
    def language_version(cls, **kwargs):  # pragma: windows
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.call.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        out = cls.call(cls.version_flags, skip_flags=True,
                       allow_error=True, **kwargs)
        if 'Copyright' not in out:  # pragma: debug
            raise RuntimeError("Version call failed: %s" % out)
        return out.split('Copyright')[0]

    
# C Archivers
class ARArchiver(ArchiverBase):
    r"""Archiver class for ar tool."""
    toolname = 'ar'
    languages = ['c', 'c++']
    default_executable_env = 'AR'
    static_library_flag = 'rcs'
    output_key = ''
    output_first_library = True


class LibtoolArchiver(ArchiverBase):
    r"""Archiver class for libtool tool."""
    toolname = 'libtool'
    languages = ['c', 'c++']
    default_executable_env = 'LIBTOOL'
    static_library_flag = '-static'  # This is the default
    

class MSVCArchiver(ArchiverBase):
    r"""Microsoft Visual Studio C Archiver."""
    toolname = 'LIB'
    languages = ['c', 'c++']
    platforms = ['Windows']
    static_library_flag = None
    output_key = '/OUT:%s'
    

# _top_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '../'))
# _incl_interface = os.path.join(_top_dir, 'interface')
_top_lang_dir = get_language_dir('c')
_incl_interface = _top_lang_dir
_incl_seri = os.path.join(_top_lang_dir, 'serialize')
_incl_comm = os.path.join(_top_lang_dir, 'communication')


class CModelDriver(CompiledModelDriver):
    r"""Class for running C models."""

    _schema_subtype_description = ('Model is written in C.')
    language = 'c'
    language_ext = ['.c', '.h']
    interface_library = 'ygg'
    supported_comms = ['ipc', 'zmq']
    supported_comm_options = {
        'ipc': {'platforms': ['MacOS', 'Linux']},
        'zmq': {'libraries': ['zmq', 'czmq']}}
    interface_directories = [_incl_interface]
    external_libraries = {
        'rapidjson': {'include': os.path.join(os.path.dirname(tools.__file__),
                                              'rapidjson', 'include',
                                              'rapidjson', 'rapidjson.h'),
                      'libtype': 'header_only',
                      'language': 'c'},
        'zmq': {'include': 'zmq.h',
                'libtype': 'shared',
                'language': 'c'},
        'czmq': {'include': 'czmq.h',
                 'libtype': 'shared',
                 'language': 'c'}}
    internal_libraries = {
        'ygg': {'source': os.path.join(_incl_interface, 'YggInterface.c'),
                # 'directory': _incl_interface,
                'linker_language': 'c++',  # Some dependencies are C++
                'internal_dependencies': ['datatypes', 'regex'],
                'external_dependencies': ['rapidjson'],
                'include_dirs': [_incl_comm, _incl_seri],
                'compiler_flags': []},
        'regex_win32': {'source': 'regex_win32.cpp',
                        'directory': os.path.join(_top_lang_dir, 'regex'),
                        'language': 'c++',
                        'libtype': _default_internal_libtype,
                        'internal_dependencies': [],
                        'external_dependencies': []},
        'regex_posix': {'source': 'regex_posix.h',
                        'directory': os.path.join(_top_lang_dir, 'regex'),
                        'language': 'c',
                        'libtype': 'header_only',
                        'internal_dependencies': [],
                        'external_dependencies': []},
        'datatypes': {'directory': os.path.join(_top_lang_dir, 'datatypes'),
                      'language': 'c++',
                      'libtype': _default_internal_libtype,
                      'internal_dependencies': ['regex'],
                      'external_dependencies': ['rapidjson'],
                      'include_dirs': []}}
    type_map = {
        'int': 'intX_t',
        'float': 'floatX_t',
        'string': 'char*',
        'array': 'vector_t',
        'object': 'map_t',
        'boolean': 'bool',
        'null': 'NULL',
        'uint': 'uintX_t',
        'complex': 'complex_X',
        'bytes': 'char*',
        'unicode': 'char*',
        '1darray': '*',
        'ndarray': '*',
        'ply': 'ply_t',
        'obj': 'obj_t',
        'schema': 'map_t'}
    function_param = {
        'comment': '//',
        'true': '1',
        'indent': 2 * ' ',
        'print': 'printf(\"{message}\");',
        'error': 'return -1;',
        'block_end': '}',
        'if_begin': 'if ({cond}) {{',
        'for_begin': ('for ({iter_var} = {iter_begin}; {iter_var} < {iter_end}; '
                      '{iter_var}++) {{'),
        'while_begin': 'while ({cond}) {{',
        'break': 'break;',
        'declare': '{type} {name};',
        'assign': '{name} = {value};',
        'exec_begin': 'int main() {',
        'exec_end': '  return 0;\n}'}

    @staticmethod
    def after_registration(cls):
        r"""Operations that should be performed to modify class attributes after
        registration."""
        if cls.default_compiler is None:
            if platform._is_linux:
                cls.default_compiler = 'gcc'
            elif platform._is_mac:
                cls.default_compiler = 'clang'
            elif platform._is_win:  # pragma: windows
                cls.default_compiler = 'cl'
        CompiledModelDriver.after_registration(cls)
        archiver = cls.get_tool('archiver')
        linker = cls.get_tool('linker')
        for x in ['zmq', 'czmq']:
            if x in cls.external_libraries:
                if platform._is_win:  # pragma: windows
                    cls.external_libraries[x]['libtype'] = 'static'
                libtype = cls.external_libraries[x]['libtype']
                if libtype == 'static':  # pragma: debug
                    tool = archiver
                    kwargs = {}
                else:
                    tool = linker
                    kwargs = {'build_library': True}
                cls.external_libraries[x][libtype] = tool.get_output_file(
                    x, **kwargs)
        # Platform specific regex internal library
        if platform._is_win:  # pragma: windows
            regex_lib = cls.internal_libraries['regex_win32']
        else:
            regex_lib = cls.internal_libraries['regex_posix']
        cls.internal_libraries['regex'] = regex_lib
        # Platform specific internal library options
        cls.internal_libraries['ygg']['include_dirs'] += [_top_lang_dir]
        if platform._is_win:  # pragma: windows
            stdint_win = os.path.join(_top_lang_dir, 'windows_stdint.h')
            assert(os.path.isfile(stdint_win))
            shutil.copy(stdint_win, os.path.join(_top_lang_dir, 'stdint.h'))
            cls.internal_libraries['datatypes']['include_dirs'] += [_top_lang_dir]
        if platform._is_linux:
            for x in ['ygg', 'datatypes']:
                if 'compiler_flags' not in cls.internal_libraries[x]:
                    cls.internal_libraries[x]['compiler_flags'] = []
                cls.internal_libraries[x]['compiler_flags'].append('-fPIC')
        
    @classmethod
    def configure(cls, cfg):
        r"""Add configuration options for this language. This includes locating
        any required external libraries and setting option defaults.

        Args:
            cfg (YggConfigParser): Config class that options should be set for.

        Returns:
            list: Section, option, description tuples for options that could not
                be set.

        """
        # Call __func__ to avoid direct invoking of class which dosn't exist
        # in after_registration where this is called
        out = CompiledModelDriver.configure.__func__(cls, cfg)
        # Change configuration to be directory containing include files
        rjlib = cfg.get(cls._language, 'rapidjson_include', None)
        if (rjlib is not None) and os.path.isfile(rjlib):
            cfg.set(cls._language, 'rapidjson_include',
                    os.path.dirname(os.path.dirname(rjlib)))
        return out
        
    @classmethod
    def update_ld_library_path(cls, env):
        r"""Update provided dictionary of environment variables so that
        LD_LIBRARY_PATH includes the interface directory containing the interface
        libraries.

        Args:
            env (dict): Dictionary of enviroment variables to be updated.

        Returns:
            dict: Updated dictionary of environment variables.

        """
        if platform._is_linux:
            path_list = []
            prev_path = env.pop('LD_LIBRARY_PATH', '')
            if prev_path:
                path_list.append(prev_path)
            for x in [cls.get_language_dir()]:
                if x not in prev_path:
                    path_list.append(x)
            if path_list:
                env['LD_LIBRARY_PATH'] = os.pathsep.join(path_list)
        return env

    def set_env(self, **kwargs):
        r"""Get environment variables that should be set for the model process.

        Args:
            **kwargs: Additional keyword arguments are passed to the parent
                class's method.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(CModelDriver, self).set_env(**kwargs)
        out = self.update_ld_library_path(out)
        return out



from __future__ import print_function
import sys
import numpy as np
from yggdrasil.interface.YggInterface import YggRpcServer
from yggdrasil.tools import sleep


def fibServer(args):

    sleeptime = float(args[0])
    print('Hello from Python rpcFibSrv: sleeptime = %f' % sleeptime)

    # Create server-side rpc conneciton using model name
    rpc = YggRpcServer("rpcFibSrv", "%d", "%d %d")

    # Continue receiving requests until error occurs (the connection is closed
    # by all clients that have connected).
    while True:
        print('rpcFibSrv(P): receiving...')
        retval, rpc_in = rpc.rpcRecv()
        if not retval:
            print('rpcFibSrv(P): end of input')
            break

        # Compute fibonacci number
        print('rpcFibSrv(P): <- input %d' % rpc_in[0], end='')
        pprev = 0
        prev = 1
        result = 1
        fib_no = 1
        arg = rpc_in[0]
        while fib_no < arg:
            result = prev + pprev
            pprev = prev
            prev = result
            fib_no = fib_no + 1
        print(' ::: ->(%2d %2d)' % (arg, result))

        # Sleep and then send response back
        sleep(float(sleeptime))
        flag = rpc.rpcSend(arg, np.int32(result))
        if not flag:
            raise RuntimeError('rpcFibSrv(P): ERROR sending')

    print('Goodbye from Python rpcFibSrv')

    
if __name__ == '__main__':
    fibServer(sys.argv[1:])



r"""This package provides a framework for integrating models across languages
such that they can be run simultaneously, passing input back and forth."""
from yggdrasil import platform
import os
import sys
import glob
import logging
import subprocess
import importlib
from ._version import get_versions
_test_package_name = None
_test_package = None
logger = logging.getLogger(__name__)
order = ['pytest', 'nose']
# order = ['nose', 'pytest']
try:
    _test_package_name = order[0]
    _test_package = importlib.import_module(_test_package_name)
except ImportError:  # pragma: debug
    try:
        _test_package_name = order[1]
        _test_package = importlib.import_module(_test_package_name)
    except ImportError:
        _test_package_name = None
        _test_package = None
    

if platform._is_win:  # pragma: windows
    # This is required to fix crash on Windows in case of Ctrl+C
    # https://github.com/ContinuumIO/anaconda-issues/issues/905#issuecomment-232498034
    os.environ['FOR_DISABLE_CONSOLE_CTRL_HANDLER'] = 'T'


def expand_and_add(path, path_list, dir_list):  # pragma: no cover
    r"""Expand the specified path and add it's expanded forms to the provided
    list.

    Args:
        path (str): Absolute/relative path with or without regex wildcard
            expressions to expand.
        path_list (list): Existing list that expansions should be added to.
        dir_list (list): Directories that should be tried for relative paths.

    Returns:
        int: The number of expansions added to path_list.

    """
    if os.path.isabs(path):
        matches = glob.glob(path)
        path_list += matches
        return len(matches)
    # Try checking for function
    for func_sep in ['::', ':']:
        if (func_sep in path):
            prepath, mod = path.rsplit(func_sep, 1)
            nadded = expand_and_add(prepath, path_list, dir_list)
            if nadded:
                for i in range(-nadded, 0):
                    path_list[i] += '%s%s' % (func_sep, mod)
                return nadded
    # Try directory prefixes
    for path_prefix in dir_list:
        nadded = expand_and_add(os.path.join(path_prefix, path),
                                path_list, dir_list)
        if nadded:
            return nadded
    return 0


def run_tsts(verbose=True, nocapture=True, stop=True,
             nologcapture=True, withcoverage=True):  # pragma: no cover
    r"""Run tests for the package. Relative paths are interpreted to be
    relative to the package root directory.

    Args:
        verbose (bool, optional): If True, set option '-v' which
            increases the verbosity. Defaults to True.
        nocapture (bool, optional): If True, set option '--nocapture'
            ('--capture=no' with pytest) which allows messages to be printed to
            stdout. Defaults to True.
        stop (bool, optional): If True, set option '--stop' ('--exitfirst' for
            pytest) which stops tests at the first failure. Defaults to True.
        nologcapture (bool, optional): If True, set option '--nologcapture'
            which allows logged messages to be printed. Defaults to True.
        withcoverage (bool, optional): If True, set option '--with-coverage'
            which invokes coverage. Defaults to True.

    """
    if _test_package is None:
        raise RuntimeError("Could not locate test runner pytest or nose.")
    elif _test_package_name == 'pytest':
        test_cmd = 'pytest'
    elif _test_package_name == 'nose':
        test_cmd = 'nosetests'
    else:
        raise RuntimeError("Unsupported test package: '%s'" % _test_package_name)
    initial_dir = os.getcwd()
    package_dir = os.path.dirname(os.path.abspath(__file__))
    error_code = 0
    argv = [test_cmd]
    test_paths = []
    opt_val = 0
    for x in sys.argv:
        if opt_val > 0:
            argv.append(x)
            opt_val -= 1
        elif x.endswith('yggtest'):
            # if _test_package_name == 'nose':
            #     argv.append(x)
            pass
        elif x == '--nocover':
            withcoverage = False
        elif x.startswith('-'):
            argv.append(x)
            if (_test_package_name == 'pytest') and (x in ['-c']):
                opt_val = 1
        else:
            test_paths.append(x)
    if _test_package_name == 'nose':
        argv += ['--detailed-errors', '--exe']
    # elif _test_package_name == 'pytest':
    #     argv.append('--ignore=yggdrasil/rapidjson/')
    if verbose:
        argv.append('-v')
    if nocapture:
        argv.append('-s')
    if stop:
        argv.append('-x')
    if nologcapture and (_test_package_name == 'nose'):
        argv.append('--nologcapture')
    if withcoverage:
        if _test_package_name == 'nose':
            argv.append('--with-coverage')
            argv.append('--cover-package=yggdrasil')
        elif _test_package_name == 'pytest':
            argv.append('--cov=%s' % package_dir)
    # Get expanded tests
    expanded_test_paths = []
    if not test_paths:
        expanded_test_paths.append(package_dir)
    else:
        for x in test_paths:
            if not expand_and_add(x, expanded_test_paths,
                                  [package_dir, os.getcwd()]):
                expanded_test_paths.append(x)
    # os.chdir(package_dir)
    argv += expanded_test_paths
    logger.info("Running %s from %s", argv, os.getcwd())
    try:
        # Set env
        old_skip_norm = os.environ.get('YGG_SKIP_COMPONENT_VALIDATION', None)
        if old_skip_norm is None:
            os.environ['YGG_SKIP_COMPONENT_VALIDATION'] = 'True'
        error_code = subprocess.call(argv)
        # if _test_package_name == 'nose':
        #     result = _test_package.run(argv=argv)
        #     if not result:
        #         error_code = -1
        # elif _test_package_name == 'pytest':
        #     error_code = _test_package.main(argv)
        # else:
        #     raise RuntimeError("No test runner.")
    except BaseException:
        logger.exception('Error in running test.')
        error_code = -1
    finally:
        os.chdir(initial_dir)
        if old_skip_norm is None:
            del os.environ['YGG_SKIP_COMPONENT_VALIDATION']
    return error_code


__all__ = []
__version__ = get_versions()['version']
del get_versions



from yggdrasil.tests import (
    assert_raises, assert_equal, assert_not_equal)
from yggdrasil.metaschema import properties, get_metaschema, _base_validator
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


existing = properties.get_registered_properties()
non_existant = 'invalid_xyz123'
existing_class = list(existing.keys())[0]
existing_validator = non_existant
assert(non_existant not in existing)
assert(non_existant not in _base_validator.VALIDATORS)
for k, v in _base_validator.VALIDATORS.items():
    if k not in existing:
        existing_validator = k
        break
else:  # pragma: debug
    # Not sure how this would happen, but guarantee it dosn't do it silently
    raise Exception('Could not locate unregistered jsonschema property.')


def test_register_metaschema_property():
    r"""Test errors in register_metaschema_property."""
    # Error when property class already registered
    args = ('ReplacementClassSchema', (MetaschemaProperty, ),
            {'name': existing_class})
    assert_raises(ValueError, type, *args)
    # Error when replacement class has schema
    args = ('ReplacementClassSchema', (MetaschemaProperty, ),
            {'name': existing_validator, 'schema': {}})
    assert_raises(ValueError, type, *args)
    # Error when validate set

    def fake_validate(*args, **kwargs):  # pragma: no cover
        return

    args = ('ReplacementClassSchema', (MetaschemaProperty, ),
            {'name': existing_validator, '_validate': fake_validate})
    assert_raises(ValueError, type, *args)
    args = ('ReplacementClassSchema', (MetaschemaProperty, ),
            {'name': existing_validator, 'schema': {}})
    assert_raises(ValueError, type, *args)
    # Error when property not in existing metaschema
    get_metaschema()  # ensures it has been initialized
    args = ('ReplacementClassSchema', (MetaschemaProperty, ),
            {'name': non_existant})
    assert_raises(ValueError, type, *args)


def test_get_registered_properties():
    r"""Test get_registered_properties."""
    assert(properties.get_registered_properties())


def test_get_metaschema_property():
    r"""Test get_metaschema_property."""
    assert_equal(properties.get_metaschema_property(non_existant),
                 MetaschemaProperty)
    assert_not_equal(properties.get_metaschema_property(existing_class),
                     MetaschemaProperty)



import numpy as np
from yggdrasil.metaschema.properties.tests import (
    test_MetaschemaProperty as parent)


class TestTypeMetaschemaProperty(parent.TestMetaschemaProperty):
    r"""Test class for TypeMetaschemaProperty class."""
    
    _mod = 'TypeMetaschemaProperty'
    _cls = 'TypeMetaschemaProperty'
    
    def __init__(self, *args, **kwargs):
        super(TestTypeMetaschemaProperty, self).__init__(*args, **kwargs)
        self._valid = [(np.int8(1), 'int'), (np.int8(1), 'scalar')]
        self._invalid = [(np.int8(1), 'float'), (np.float32(1), 'int')]
        self._encode_errors = [np]  # Can't encode modules
        self._valid_compare = [('int', 'int'), ('int', 'scalar'),
                               ('ply', 'object')]
        self._invalid_compare = [('int', 'float'), ('array', 'object'),
                                 ('ply', 'array'), ('1darray', 'scalar')]



import os
import sys
import warnings
from setuptools import setup, find_packages
from distutils.sysconfig import get_python_lib
import versioneer
import install_matlab_engine
import create_coveragerc
ygg_ver = versioneer.get_version()
ROOT_PATH = os.path.abspath(os.path.dirname(__file__))


# Attempt to install openalea
try:
    from openalea import lpy
    lpy_installed = True
except ImportError:
    warnings.warn("Could not import openalea.lpy. " +
                  "LPy support will be disabled.")
    lpy_installed = False


# Attempt to install matlab engine
if install_matlab_engine.install_matlab(as_user=('--user' in sys.argv)):
    matlab_installed = True
else:
    warnings.warn("Could not import matlab.engine. " +
                  "Matlab features will be disabled.")
    matlab_installed = False


# Determine if rapidjson installed and parse user defined location
rj_include_dir0 = os.path.join(ROOT_PATH, 'yggdrasil', 'rapidjson', 'include')
for idx, arg in enumerate(sys.argv[:]):
    if ((arg.startswith('--rj-include-dir=')
         or arg.startswith('--rapidjson-include-dir='))):
        sys.argv.pop(idx)
        rj_include_dir = os.path.abspath(arg.split('=', 1)[1])
        break
else:
    rj_include_dir = rj_include_dir0
if not os.path.isdir(rj_include_dir):
    raise RuntimeError("RapidJSON sources could not be located. If you "
                       "cloned the git repository, initialize the rapidjson "
                       "git submodule by calling "
                       "'git submodule update --init --recursive' "
                       "from inside the repository.")
if rj_include_dir != rj_include_dir0:
    def_config_file = os.path.join(ROOT_PATH, 'yggdrasil', 'defaults.cfg')
    try:
        import ConfigParser as configparser
    except ImportError:
        import configparser
    cfg = configparser.ConfigParser()
    cfg.read(def_config_file)
    if not cfg.has_section('c'):
        cfg.add_section('c')
    cfg.set('c', 'rapidjson_include', rj_include_dir)
    with open(def_config_file, 'w') as fd:
        cfg.write(fd)

    
# Set coverage options in .coveragerc
create_coveragerc.create_coveragerc(matlab_installed=matlab_installed,
                                    lpy_installed=lpy_installed)


# Create .rst README from .md and get long description
if os.path.isfile('README.rst'):
    with open('README.rst', 'r') as file:
        long_description = file.read()
elif os.path.isfile('README.md'):
    try:
        import pypandoc
        pypandoc.convert_file('README.md', 'rst', outputfile='README.rst')
        long_description = pypandoc.convert_file('README.md', 'rst')
    except (ImportError, IOError):
        with open('README.md', 'r') as file:
            long_description = file.read()
else:
    raise IOError("Could not find README.rst or README.md")


# Create requirements list based on platform
requirements = ['numpy>=1.13.0', "scipy", "pyyaml",
                "pystache", "pyzmq", "psutil",
                "matplotlib<3.0; python_version < '3.5'",
                "matplotlib; python_version >= '3.5'",
                "jsonschema",
                "python-rapidjson; python_version >= '3.4'",
                'pandas<0.21; python_version == "3.4"',
                'pandas; python_version != "3.4"',
                "perf",
                "pint; python_version == '2.7'",
                "unyt; python_version >= '3.4'",
                "six",
                'sysv_ipc; platform_system != "Windows"']
test_requirements = ['pytest']
# optional_requirements = ["pika<1.0", "astropy"]


# Warn that local install may not have entry points on path
if '--user' in sys.argv:
    script_dir = os.path.realpath(os.path.join(get_python_lib(),
                                               '../../../bin/'))
    warnings.warn("When installing locally, you may need to add the script " +
                  "directory to your path manually in order to have access " +
                  "to the command line entry points (e.g. yggrun). " +
                  "If 'yggrun' is not a recognized command, try adding " +
                  "'%s' to your PATH." % script_dir)
    
setup(
    name="yggdrasil-framework",
    packages=find_packages(),
    include_package_data=True,
    version=ygg_ver,
    cmdclass=versioneer.get_cmdclass(),
    description=("A framework for combining interdependent models from "
                 "multiple languages."),
    long_description=long_description,
    author="Meagan Lang",
    author_email="langmm.astro@gmail.com",
    url="https://github.com/cropsinsilico/yggdrasil",
    download_url=(
        "https://github.com/cropsinsilico/yggdrasil/archive/%s.tar.gz" % ygg_ver),
    keywords=["plants", "simulation", "models", "framework"],
    install_requires=requirements,
    tests_require=test_requirements,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: C",
        "Programming Language :: C++",
        "Programming Language :: ML",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering",
        "Development Status :: 3 - Alpha",
    ],
    entry_points={
        'console_scripts': ['yggrun=yggdrasil.command_line:yggrun',
                            'cisrun=yggdrasil.command_line:yggrun',
                            'yggcc=yggdrasil.command_line:yggcc',
                            'yggccflags=yggdrasil.command_line:cc_flags',
                            'yggldflags=yggdrasil.command_line:ld_flags',
                            'yggtest=yggdrasil:run_tsts',
                            'yggmetaschema=yggdrasil.command_line:regen_metaschema',
                            'yggschema=yggdrasil.command_line:regen_schema',
                            'yggbuildapi_c=yggdrasil.command_line:rebuild_c_api',
                            'yggconfig=yggdrasil.command_line:update_config',
                            'yggtime_comm=yggdrasil.command_line:yggtime_comm',
                            'yggtime_lang=yggdrasil.command_line:yggtime_lang',
                            'yggtime_os=yggdrasil.command_line:yggtime_os',
                            'yggtime_py=yggdrasil.command_line:yggtime_py',
                            'yggtime_paper=yggdrasil.command_line:yggtime_paper',
                            'yggvalidate=yggdrasil.command_line:validate_yaml'],
    },
    license="BSD",
)



import subprocess
import uuid as uuid_gen
import logging
from datetime import datetime
import os
import psutil
import warnings
import weakref
from yggdrasil import backwards, tools, platform
from yggdrasil.config import ygg_cfg
from yggdrasil.drivers.InterpretedModelDriver import InterpretedModelDriver
from yggdrasil.tools import TimeOut, sleep
from yggdrasil.languages import get_language_dir
logger = logging.getLogger(__name__)
try:  # pragma: matlab
    disable_engine = ygg_cfg.get('matlab', 'disable_engine', 'False').lower()
    if platform._is_win or (disable_engine == 'true'):
        _matlab_engine_installed = False
        if not tools.is_subprocess():
            logger.debug("matlab.engine disabled")
    else:
        import matlab.engine
        _matlab_engine_installed = True
except ImportError:  # pragma: no matlab
    logger.debug("Could not import matlab.engine. "
                 + "Matlab support for using a sharedEngine will be disabled.")
    _matlab_engine_installed = False


_top_dir = os.path.normpath(os.path.join(os.path.dirname(__file__), '../'))
_top_lang_dir = get_language_dir('matlab')
_compat_map = {
    'R2015b': ['2.7', '3.3', '3.4'],
    'R2017a': ['2.7', '3.3', '3.4', '3.5'],
    'R2017b': ['2.7', '3.3', '3.4', '3.5', '3.6'],
    'R2018b': ['2.7', '3.3', '3.4', '3.5', '3.6']}


def kill_all():
    r"""Kill all Matlab shared engines."""
    if platform._is_win:  # pragma: windows
        os.system(('taskkill /F /IM matlab.engine.shareEngine /T'))
    else:
        os.system(('pkill -f matlab.engine.shareEngine'))


def locate_matlab_engine_processes():  # pragma: matlab
    r"""Get all of the active matlab sharedEngine processes.

    Returns:
        list: Active matlab sharedEngine processes.

    """
    out = []
    for p in psutil.process_iter():
        p.info = p.as_dict(attrs=['name', 'pid', 'cmdline'])
        if (((p.info['name'] == 'MATLAB')
             and ('matlab.engine.shareEngine' in p.info['cmdline']))):
            out.append(p)  # p.info['pid'])
    return out


def is_matlab_running():
    r"""Determine if there is a Matlab engine running.

    Returns:
        bool: True if there is a Matlab engine running, False otherwise.

    """
    if not _matlab_engine_installed:  # pragma: no matlab
        out = False
    else:  # pragma: matlab
        out = (len(matlab.engine.find_matlab()) != 0)
    return out


def locate_matlabroot():  # pragma: matlab
    r"""Find directory that servers as matlab root.

    Returns:
        str: Full path to matlabroot directory.

    """
    return MatlabModelDriver.get_matlab_info()[0]


def install_matlab_engine():  # pragma: matlab
    r"""Install the MATLAB engine API for Python."""
    if not _matlab_engine_installed:
        mtl_root = locate_matlabroot()
        mtl_setup = os.path.join(mtl_root, 'extern', 'engines', 'python')
        cmd = 'python setup.py install'
        result = subprocess.check_output(cmd, cwd=mtl_setup)
        print(result)
    

def start_matlab_engine(skip_connect=False, timeout=None):  # pragma: matlab
    r"""Start a Matlab shared engine session inside a detached screen
    session.

    Args:
        skip_connect (bool, optional): If True, the engine is not connected.
            Defaults to False.
        timeout (int, optional): Time (in seconds) that should be waited for
            Matlab to start up. Defaults to None and is set from the config
            option ('matlab', 'startup_waittime_s').

    Returns:
        tuple: Information on the started session including the name of the
            screen session running matlab, the created engine object, the name
            of the matlab session, and the matlab engine process.

    Raises:
        RuntimeError: If Matlab is not installed.

    """
    if not _matlab_engine_installed:  # pragma: no matlab
        raise RuntimeError("Matlab engine is not installed.")
    if timeout is None:
        timeout = float(ygg_cfg.get('matlab', 'startup_waittime_s', 10))
    old_process = set(locate_matlab_engine_processes())
    old_matlab = set(matlab.engine.find_matlab())
    screen_session = str('ygg_matlab' + datetime.today().strftime("%Y%j%H%M%S")
                         + '_%d' % len(old_matlab))
    try:
        args = ['screen', '-dmS', screen_session, '-c',
                os.path.join(_top_lang_dir, 'matlab_screenrc'),
                'matlab', '-nodisplay', '-nosplash', '-nodesktop', '-nojvm',
                '-r', '"matlab.engine.shareEngine"']
        subprocess.call(' '.join(args), shell=True)
        T = TimeOut(timeout)
        while ((len(set(matlab.engine.find_matlab()) - old_matlab) == 0)
               and not T.is_out):
            logger.debug('Waiting for matlab engine to start')
            sleep(1)  # Usually 3 seconds
    except KeyboardInterrupt:  # pragma: debug
        args = ['screen', '-X', '-S', screen_session, 'quit']
        subprocess.call(' '.join(args), shell=True)
        raise
    if (len(set(matlab.engine.find_matlab()) - old_matlab) == 0):  # pragma: debug
        raise Exception("start_matlab timed out at %f s" % T.elapsed)
    new_matlab = list(set(matlab.engine.find_matlab()) - old_matlab)[0]
    new_process = list(set(locate_matlab_engine_processes()) - old_process)[0]
    # Connect to the engine
    matlab_engine = None
    if not skip_connect:
        matlab_engine = connect_matlab_engine(new_matlab, first_connect=True)
    return screen_session, matlab_engine, new_matlab, new_process


def connect_matlab_engine(matlab_session, first_connect=False):  # pragma: matlab
    r"""Connect to Matlab engine.

    Args:
        matlab_session (str): Name of the Matlab session that should be
            connected to.
        first_connect (bool, optional): If True, this is the first time
            Python is connecting to the Matlab shared engine and certain
            environment variables should be set. Defaults to False.

    Returns:
        MatlabEngine: Matlab engine that was connected.

    """
    matlab_engine = matlab.engine.connect_matlab(matlab_session)
    matlab_engine.eval('clear classes;', nargout=0)
    err = backwards.StringIO()
    try:
        matlab_engine.eval("YggInterface('YGG_MSG_MAX');", nargout=0,
                           stderr=err)
    except BaseException:
        matlab_engine.addpath(_top_dir, nargout=0)
        matlab_engine.addpath(_top_lang_dir, nargout=0)
    matlab_engine.eval("os = py.importlib.import_module('os');", nargout=0)
    if not first_connect:
        if backwards.PY2:
            matlab_engine.eval("py.reload(os);", nargout=0)
        else:
            # matlab_engine.eval("py.importlib.reload(os);", nargout=0)
            pass
    return matlab_engine


def stop_matlab_engine(screen_session, matlab_engine, matlab_session,
                       matlab_process, keep_engine=False):  # pragma: matlab
    r"""Stop a Matlab shared engine session running inside a detached screen
    session.

    Args:
        screen_session (str): Name of the screen session that the shared
            Matlab session was started in.
        matlab_engine (MatlabEngine): Matlab engine that should be stopped.
        matlab_session (str): Name of Matlab session that the Matlab engine is
            connected to.
        matlab_process (psutil.Process): Process running the Matlab shared engine.
        keep_engine (bool, optional): If True, the references to the engine will be
            removed so it is not deleted. Defaults to False.

    Raises:
        RuntimeError: If Matlab is not installed.

    """
    if not _matlab_engine_installed:  # pragma: no matlab
        raise RuntimeError("Matlab engine is not installed.")
    if keep_engine and (matlab_engine is not None):
        if '_matlab' in matlab_engine.__dict__:
            matlab_engine.quit()
        return
    # Remove weakrefs to engine to prevent stopping engine more than once
    if matlab_engine is not None:
        # Remove weak references so engine not deleted on exit
        eng_ref = weakref.getweakrefs(matlab_engine)
        for x in eng_ref:
            if x in matlab.engine._engines:
                matlab.engine._engines.remove(x)
        # Either exit the engine or remove its reference
        if matlab_session in matlab.engine.find_matlab():
            try:
                matlab_engine.eval('exit', nargout=0)
            except BaseException:
                pass
        else:  # pragma: no cover
            matlab_engine.__dict__.pop('_matlab', None)
    # Stop the screen session containing the Matlab shared session
    if screen_session is not None:
        if matlab_session in matlab.engine.find_matlab():
            os.system(('screen -X -S %s quit') % screen_session)
        T = TimeOut(5)
        while ((matlab_session in matlab.engine.find_matlab())
               and not T.is_out):
            logger.debug("Waiting for matlab engine to exit")
            sleep(1)
        if (matlab_session in matlab.engine.find_matlab()):  # pragma: debug
            if matlab_process is not None:
                matlab_process.terminate()
                logger.error("stop_matlab_engine timed out at %f s. " % T.elapsed
                             + "Killed Matlab sharedEngine process.")


class MatlabProcess(tools.YggClass):  # pragma: matlab
    r"""Add features to mimic subprocess.Popen while running Matlab function
    asynchronously.

    Args:
        target (func): Matlab function that should be called.
        args (list, tuple): Arguments that should be passed to target.
        kwargs (dict, optional): Keyword arguments that should be passed to
            target. Defaults to empty dict.
        name (str, optional): A name for the process. Generated if not provided.
        matlab_engine (MatlabEngine, optional): MatlabEngine that should be used
            to get errors. Defaults to None and errors will not be recovered
            unless passed through stdout and stderr before shutdown.

    Attributes:
        stdout (StringIO): File like string buffer that stdout from target will
            be written to.
        stderr (StringIO): File like string buffer that stderr from target will
            be written to.
        target (func): Matlab function that should be called.
        args (list, tuple): Arguments that should be passed to target.
        kwargs (dict): Keyword arguments that should be passed to target.
        future (MatlabFutureResult): Future result from async function. This
            will be None until start is called.
        matlab_engine (MatlabEngine): MatlabEngine that should be used to get
            errors.

    Raises:
        RuntimeError: If Matlab is not installed.

    """

    def __init__(self, target, args, kwargs=None, name=None, matlab_engine=None):
        if not _matlab_engine_installed:  # pragma: no matlab
            raise RuntimeError("Matlab engine is not installed.")
        if kwargs is None:
            kwargs = {}
        self.stdout = backwards.sio.StringIO()
        self.stderr = backwards.sio.StringIO()
        self._stdout_line = None
        self._stderr_line = None
        self.target = target
        self.args = args
        self.kwargs = kwargs
        self.kwargs.update(nargout=0, stdout=self.stdout, stderr=self.stderr)
        self.kwargs['async'] = True  # For python 3.7 where async is reserved
        self.future = None
        self.matlab_engine = matlab_engine
        self._returncode = None
        super(MatlabProcess, self).__init__(name)

    def poll(self, *args, **kwargs):
        r"""Fake poll."""
        return self.returncode

    @property
    def stdout_line(self):
        r"""str: Output to stdout from function call."""
        if self._stdout_line is None:
            if self.stdout is not None:
                line = self.stdout.getvalue()
                if line:
                    self._stdout_line = line
        return self._stdout_line

    @property
    def stderr_line(self):
        r"""str: Output to stderr from function call."""
        if self._stderr_line is None:
            if self.stderr is not None:
                line = self.stderr.getvalue()
                if line:
                    self._stderr_line = line
        return self._stderr_line

    def print_output(self):
        r"""Print output from stdout and stderr."""
        if self.stdout_line:
            self.print_encoded(self.stdout_line, end="")
        if self.stderr_line:
            self.print_encoded(self.stderr_line, end="")
            
    def start(self):
        r"""Start asychronous call."""
        self.future = self.target(*self.args, **self.kwargs)

    def is_started(self):
        r"""bool: Has start been called."""
        return (self.future is not None)

    def is_cancelled(self):
        r"""bool: Was the async call cancelled or not."""
        if self.is_started():
            try:
                return self.future.cancelled()
            except matlab.engine.EngineError:
                self.on_matlab_error()
                return True
            except BaseException:
                return True
        return False

    def is_done(self):
        r"""bool: Is the async call still running."""
        if self.is_started():
            try:
                return self.future.done() or self.is_cancelled()
            except matlab.engine.EngineError:
                self.on_matlab_error()
                return True
            except BaseException:
                return True
        return False

    def is_alive(self):
        r"""bool: Is the async call funning."""
        if self.is_started():
            return (not self.is_done())
        return False

    @property
    def returncode(self):
        r"""int: Return code."""
        if self.is_done():
            if self.stderr_line:  # or self.is_cancelled():
                return -1
            else:
                return 0
        else:
            return self._returncode

    def kill(self, *args, **kwargs):
        r"""Cancel the async call."""
        if self.is_alive():
            try:
                out = self.future.cancel()
                self.debug("Result of cancelling Matlab call?: %s", out)
            except matlab.engine.EngineError as e:
                self.debug('Matlab Engine Error: %s' % e)
                self.on_matlab_error()
            except BaseException as e:
                self.debug('Other error on kill: %s' % e)
        self.print_output()
        if self.is_alive():
            self.info('Error killing Matlab script.')
            self.matlab_engine.quit()
            self.future = None
            self._returncode = -1
        assert(not self.is_alive())

    def on_matlab_error(self):
        r"""Actions performed on error in Matlab engine."""
        # self.print_output()
        self.debug('')
        if self.matlab_engine is not None:
            try:
                self.matlab_engine.eval('exception = MException.last;', nargout=0)
                self.matlab_engine.eval('getReport(exception)')
            except matlab.engine.EngineError:
                pass


class MatlabModelDriver(InterpretedModelDriver):  # pragma: matlab
    r"""Base class for running Matlab models.

    Args:
        name (str): Driver name.
        args (str or list): Argument(s) for running the model in matlab.
            Generally, this should be the full path to a Matlab script.
        **kwargs: Additional keyword arguments are passed to parent class's
            __init__ method.

    Attributes:
        started_matlab (bool): True if the driver had to start a new matlab
            engine. False otherwise.
        screen_session (str): Screen session that Matlab was started in.
        mlengine (object): Matlab engine used to run script.
        mlsession (str): Name of the Matlab session that was started.

    Raises:
        RuntimeError: If Matlab is not installed.

    .. note:: Matlab models that call exit will shut down the shared engine.

    """

    _schema_subtype_description = ('Model is written in Matlab.')
    language = 'matlab'
    language_ext = '.m'
    base_languages = ['python']
    default_interpreter_flags = ['-nodisplay', '-nosplash', '-nodesktop',
                                 '-nojvm', '-batch']
    version_flags = ["fprintf('R%s', version('-release')); exit();"]
    type_map = {
        'int': 'intX',
        'float': 'single, double',
        'string': 'char',
        'array': 'cell',
        'object': 'containers.Map',
        'boolean': 'logical',
        'null': 'NaN',
        'uint': 'uintX',
        'complex': 'complex',
        'bytes': 'char (utf-8)',
        'unicode': 'char',
        '1darray': 'mat',
        'ndarray': 'mat',
        'ply': 'containers.Map',
        'obj': 'containers.Map',
        'schema': 'containers.Map'}
    function_param = {
        'comment': '%',
        'true': 'true',
        'indent': 2 * ' ',
        'print': 'disp(\'{message}\');',
        'error': 'error(\'{error_msg}\');',
        'block_end': 'end;',
        'if_begin': 'if ({cond})',
        'for_begin': 'for {iter_var} = {iter_begin}:{iter_end}',
        'while_begin': 'while ({cond})',
        'try_begin': 'try',
        'try_except': 'catch {error_var}',
        'assign': '{name} = {value};'}

    def __init__(self, name, args, **kwargs):
        self.using_matlab_engine = _matlab_engine_installed
        if self.using_matlab_engine:
            kwargs['skip_interpreter'] = True
        self.model_wrapper = None
        super(MatlabModelDriver, self).__init__(name, args, **kwargs)
        self.started_matlab = False
        self.screen_session = None
        self.mlengine = None
        self.mlsession = None
        self.mlprocess = None

    def parse_arguments(self, args):
        r"""Sort model arguments to determine which one is the executable
        and which ones are arguments.

        Args:
            args (list): List of arguments provided.

        """
        super(MatlabModelDriver, self).parse_arguments(args)
        model_base, model_ext = os.path.splitext(os.path.basename(self.model_file))
        wrap_base = 'wrapped_%s_%s' % (model_base, self.uuid.replace('-', '_'))
        # Matlab has a variable name limit of 62
        wrap_base = wrap_base[:min(len(wrap_base), 60)]
        self.model_wrapper = os.path.join(self.model_dir, wrap_base + model_ext)
        self.wrapper_products.append(self.model_wrapper)
        
    @classmethod
    def write_error_wrapper(cls, fname, try_lines, matlab_engine=None):
        r"""Write a wrapper for the model that encloses it in a try except so
        that the error can be propagated appropriately.

        Args:
            fname (str): File where the wrapper should be written.
            try_lines (list): List of lines to go in the try block.
            model_file (str): Path to model that should be wrapped.
            matlab_engine (MatlabEngine, optional): Matlab engine that will be
               used to call the wrapper. If not provided, it is assumed the
               error will be called using the Matlab interpreter on the command
               line. Defautls to None.

        Raises:

        """
        # Create lines based on use of engine or not
        if matlab_engine is not None:
            catch_block = ["error(e.message);"]
        else:
            catch_block = ["rethrow(e);"]
            # catch_block = ["fprintf('MATLAB ERROR:\\n%s\\n', e.message);",
            #                "disp(e.identifier);",
            #                "disp(e.stack);",
            #                "exit(0);"]
        lines = cls.write_try_except(try_lines, catch_block)
        if matlab_engine is None:
            lines.append("exit(0);")
        # Write lines
        logger.debug('Wrapper:\n\t%s', '\n\t'.join(lines))
        if fname is None:
            return lines
        else:
            if os.path.isfile(fname):  # pragma: debug
                os.remove(fname)
            with open(fname, 'w') as fd:
                fd.write('\n'.join(lines))
            logger.debug("Wrote wrapper to: %s" % fname)

    @classmethod
    def run_executable(cls, args, dont_wrap_error=False, fname_wrapper=None,
                       matlab_engine=None, **kwargs):
        r"""Run a program using the executable for this language and the
        provided arguments.

        Args:
            args (list): The program that should be run and any arguments
                that should be provided to it.
            dont_wrap_error (bool, optional): If False, the executable will be
                wrapped in a try/catch block to prevent errors from stopping
                Matlab shutdown. If True, the command will be executed as is
                with the Matlab interpreter. Defaults to False.
            fname_wrapper (str, optional): File where wrapper should be saved.
                If not provided, one is created. Defaults to None.
            matlab_engine (MatlabEngine, optional): Matlab engine that should be
                used to run the command. If not provided, the Matlab interpreter
                is used instead. Defaults to None.
            **kwargs: Additional keyword arguments are passed to
                cls.executable_command and tools.popen_nobuffer.

        Returns:
            str: Output to stdout from the run command.
        
        Raises:
            RuntimeError: If the language is not installed.
            RuntimeError: If there is an error when running the command.

        """
        # Strip file if first argument is a file
        if os.path.isfile(args[0]):
            kwargs.setdefault('working_dir', os.path.dirname(args[0]))
            args = [os.path.splitext(os.path.basename(args[0]))[0]] + args[1:]
        # Write wrapper
        if (not dont_wrap_error) and (len(args) > 0):
            if len(args) == 1:
                # TODO: Will this work if there is a function defined in the
                # script?
                try_block = [args[0]]
                if not try_block[0].endswith(';'):
                    try_block[0] += ';'
            else:
                # Put quotes around arguments since they would be strings when
                # passed from the command line
                func_call = "%s('%s'" % (args[0], args[1])
                for a in args[2:]:
                    func_call += (", '%s'" % a)
                func_call += ');'
                try_block = [func_call]
            if fname_wrapper is None:
                fname_wrapper = 'wrapper_%s%s' % (str(uuid_gen.uuid4()),
                                                  cls.language_ext[0])
                fname_wrapper = fname_wrapper.replace('-', '_')
                working_dir = kwargs.get('working_dir', kwargs.get('cwd', None))
                if working_dir is not None:
                    fname_wrapper = os.path.join(working_dir, fname_wrapper)
            cls.write_error_wrapper(fname_wrapper, try_block,
                                    matlab_engine=matlab_engine)
            assert(os.path.isfile(fname_wrapper))
            args = [os.path.splitext(os.path.basename(fname_wrapper))[0]]
        # Call base, catching error to remove temp wrapper
        try:
            if matlab_engine is None:
                kwargs['for_matlab'] = True
                out = super(MatlabModelDriver, cls).run_executable(args, **kwargs)
            else:
                assert(kwargs.get('return_process', False))
                # Add environment variables
                env = kwargs.get('env', {})
                old_env = {}
                new_env_str = ''
                for k, v in env.items():
                    old_env[k] = matlab_engine.getenv(k)
                    matlab_engine.setenv(k, v, nargout=0)
                    new_env_str += "'%s', %s, " % (k, repr(v))
                matlab_engine.eval('new_env = py.dict(pyargs(%s));'
                                   % new_env_str[:-2], nargout=0)
                matlab_engine.eval('os.environ.update(new_env);', nargout=0)
                # Create matlab process using Matlab engine
                out = MatlabProcess(name=args[0] + '.MatlabProcess',
                                    target=getattr(matlab_engine, args[0]),
                                    args=args[1:], matlab_engine=matlab_engine)
                out.start()
        finally:
            if (((not kwargs.get('return_process', False))
                 and (fname_wrapper is not None))):
                os.remove(fname_wrapper)
        return out
        
    @classmethod
    def language_version(cls):
        r"""Determine the version of this language.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        return cls.get_matlab_info()[1]
        
    @classmethod
    def executable_command(cls, args, **kwargs):
        r"""Compose a command for running a program in this language with the
        provied arguments. If not already present, the interpreter command and
        interpreter flags are prepended to the provided arguments.

        Args:
            args (list): The program that returned command should run and any
                arguments that should be provided to it.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Arguments composing the command required to run the program
                from the command line using the interpreter for this language.

        """
        # if kwargs.get('exec_type', 'interpreter') == 'interpreter':
        #     args = ["\"%s\"" % (' '.join(args))]
        return super(MatlabModelDriver, cls).executable_command(args, **kwargs)
    
    @classmethod
    def configure(cls, cfg):
        r"""Add configuration options for this language. This includes locating
        any required external libraries and setting option defaults.

        Args:
            cfg (YggConfigParser): Config class that options should be set for.

        Returns:
            list: Section, option, description tuples for options that could not
                be set.

        """
        out = super(MatlabModelDriver, cls).configure(cfg)
        opts = {
            'startup_waittime_s': [('The time allowed for a Matlab engine to start'
                                    'before timing out and reporting an error.'),
                                   '10'],
            'version': ['The version (release number) of installed Matlab.', ''],
            'matlabroot': ['The path to the default installation of matlab.', '']}
        if cfg.get(cls.language, 'disable', 'False').lower() != 'true':
            try:
                opts['matlabroot'][1], opts['version'][1] = cls.get_matlab_info()
            except RuntimeError:  # pragma: no matlab
                pass
        for k in opts.keys():
            if not cfg.has_option(cls.language, k):
                if opts[k][1]:  # pragma: matlab
                    cfg.set(cls.language, k, opts[k][1])
                else:
                    out.append((cls.language, k, opts[k][0]))
        return out

    @classmethod
    def get_matlab_info(cls):  # pragma: matlab
        r"""Determine the root directory where Matlab is installed and the version
        that is installed (if Matlab is installed at all). This will fail if Matlab
        is not installed, cannot be started, or does not operate as expected.

        Returns:
            tuple: Matlab root directory and Matlab version string.

        Raises:
            RuntimeError: If Matlab cannot be started or the root directory or
                release cannot be determiend.

        """
        mtl_id = '=MATLABROOT='
        cmd = ("fprintf('" + mtl_id + "%s" + mtl_id + "R%s" + mtl_id + "'"
               + ",matlabroot,version('-release'));")
        mtl_proc = cls.run_executable([cmd])
        mtl_id = backwards.match_stype(mtl_proc, mtl_id)
        if mtl_id not in mtl_proc:  # pragma: debug
            raise RuntimeError(("Could not locate ID string (%s) in "
                                "output (%s).") % (mtl_id, mtl_proc))
        parts = mtl_proc.split(mtl_id)
        if len(parts) < 3:  # pragma: debug
            raise RuntimeError(("Could not get matlabroot/version from "
                                "output (%s).") % (mtl_proc))
        matlabroot = backwards.as_str(parts[-3])
        release = backwards.as_str(parts[-2])
        return matlabroot, release

    def start_matlab_engine(self):
        r"""Start matlab session and connect to it."""
        ml_attr = ['screen_session', 'mlengine', 'mlsession', 'mlprocess']
        attempt_connect = (len(matlab.engine.find_matlab()) != 0)
        # Connect to matlab if a session exists
        if attempt_connect:
            for mlsession in matlab.engine.find_matlab():
                try:
                    self.debug("Trying to connect to session %s", mlsession)
                    self.mlengine = connect_matlab_engine(mlsession)
                    self.mlsession = mlsession
                    self.debug("Connected to existing shared engine: %s",
                               self.mlsession)
                    break
                except matlab.engine.EngineError:
                    pass
        # Start if not running or connect failed
        if self.mlengine is None:
            if attempt_connect:
                self.debug("Starting a matlab shared engine (connect failed)")
            else:
                self.debug("Starting a matlab shared engine (none existing)")
            out = start_matlab_engine()
            for i, attr in enumerate(ml_attr):
                setattr(self, attr, out[i])
            self.started_matlab = True
        # Add things to Matlab environment
        self.mlengine.addpath(self.model_dir, nargout=0)
        self.debug("Connected to matlab session '%s'" % self.mlsession)

    def set_env(self):
        r"""Get environment variables that should be set for the model process.

        Returns:
            dict: Environment variables for the model process.

        """
        out = super(MatlabModelDriver, self).set_env()
        path_list = []
        prev_path = out.pop('MATLABPATH', '')
        if prev_path:
            path_list.append(prev_path)
        for x in [_top_dir, _top_lang_dir, self.model_dir]:
            if x not in prev_path:
                path_list.append(x)
        if path_list:
            out['MATLABPATH'] = os.pathsep.join(path_list)
        return out
        
    def before_start(self):
        r"""Actions to perform before the run loop."""
        kwargs = dict(fname_wrapper=self.model_wrapper)
        if self.using_matlab_engine:
            self.start_matlab_engine()
            kwargs.update(matlab_engine=self.mlengine,
                          no_queue_thread=True)
        else:
            kwargs.update(working_dir=self.model_dir)
        with self.lock:
            if self.using_matlab_engine and (self.mlengine is None):  # pragma: debug
                self.debug('Matlab engine not set. Stopping')
                return
            super(MatlabModelDriver, self).before_start(**kwargs)

    def run_loop(self):
        r"""Loop to check if model is still running and forward output."""
        if self.using_matlab_engine:
            self.model_process.print_output()
            self.periodic_debug('matlab loop', period=100)('Looping')
            if self.model_process.is_done():
                self.model_process.print_output()
                self.set_break_flag()
                try:
                    self.model_process.future.result()
                    self.model_process.print_output()
                except matlab.engine.EngineError:
                    self.model_process.print_output()
                except BaseException:
                    self.model_process.print_output()
                    self.exception("Error running model.")
            else:
                self.sleep()
        else:
            super(MatlabModelDriver, self).run_loop()

    def after_loop(self):
        r"""Actions to perform after run_loop has finished. Mainly checking
        if there was an error and then handling it."""
        if self.using_matlab_engine:
            if (self.model_process is not None) and self.model_process.is_alive():
                self.info("Model process thread still alive")
                self.kill_process()
                return
        super(MatlabModelDriver, self).after_loop()
        if self.using_matlab_engine:
            with self.lock:
                self.cleanup()

    def cleanup(self):
        r"""Close the Matlab session and engine."""
        if self.using_matlab_engine:
            try:
                stop_matlab_engine(self.screen_session, self.mlengine,
                                   self.mlsession, self.mlprocess,
                                   keep_engine=(not self.started_matlab))
            except (SystemError, Exception) as e:  # pragma: debug
                self.error('Failed to exit matlab engine')
                self.raise_error(e)
            self.debug('Stopped Matlab')
            self.screen_session = None
            self.mlsession = None
            self.started_matlab = False
            self.mlengine = None
            self.mlprocess = None
        super(MatlabModelDriver, self).cleanup()

    def check_exits(self):
        r"""Check to make sure the program dosn't contain any exits as exits
        will shut down the Matlab engine as well as the program.

        Raises:
            RuntimeError: If there are any exit calls in the file.

        """
        has_exit = False
        with open(self.raw_model_file, 'r') as fd:
            for i, line in enumerate(fd):
                if line.strip().startswith('exit'):
                    has_exit = True
                    break
        if self.using_matlab_engine and has_exit:
            warnings.warn(
                "Line %d in '%s' contains an " % (
                    i, self.raw_model_file)
                + "'exit' call which will exit the MATLAB engine "
                + "such that it cannot be reused. Please replace 'exit' "
                + "with a return or error.")



import os
import uuid
import unittest
import tempfile
import shutil
from yggdrasil import runner, tools, platform
from yggdrasil.examples import yamls, source, ext_map
from yggdrasil.tests import YggTestBase


_ext2lang = {v: k for k, v in ext_map.items()}


class TestExample(YggTestBase, tools.YggClass):
    r"""Base class for running examples."""

    example_name = None
    expects_error = False
    env = {}

    def __init__(self, *args, **kwargs):
        tools.YggClass.__init__(self, self.example_name)
        self.language = None
        self.uuid = str(uuid.uuid4())
        self.runner = None
        # self.debug_flag = True
        super(TestExample, self).__init__(*args, **kwargs)

    @property
    def description_prefix(self):
        r"""Prefix message with test name."""
        return self.name

    @property
    def namespace(self):
        r"""str: Namespace for the example."""
        return "%s_%s" % (self.name, self.uuid)

    @property
    def tempdir(self):
        r"""str: Temporary directory."""
        return tempfile.gettempdir()

    @property
    def languages_tested(self):
        r"""list: Languages covered by the example."""
        if self.name not in source:  # pragma: debug
            return None
        if self.language not in yamls[self.name]:  # pragma: debug
            return None
        if self.language in ['all', 'all_nomatlab']:
            out = [_ext2lang[os.path.splitext(x)[-1]] for x in
                   source[self.name][self.language]]
        else:
            out = [self.language]
        return out

    @property
    def yaml(self):
        r"""str: The full path to the yaml file for this example."""
        if self.name not in yamls:
            return None
        if self.language not in yamls[self.name]:
            return None
        return yamls[self.name][self.language]

    @property
    def yamldir(self):
        r"""str: Full path to the directory containing the yaml file."""
        if self.yaml is None:  # pragma: no cover
            return None
        if isinstance(self.yaml, list):
            out = os.path.dirname(self.yaml[0])
        else:
            out = os.path.dirname(self.yaml)
        return out

    # @property
    # def yaml_contents(self):
    #     r"""dict: Contents of yaml file."""
    #     if self.yaml is None:  # pragma: no cover
    #         return None
    #     return tools.parse_yaml(self.yaml)

    @property
    def input_files(self):  # pragma: debug
        r"""list Input files for the run."""
        return None

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return None

    @property
    def results(self):
        r"""list: Results that should be found in the output files."""
        if self.input_files is None:  # pragma: debug
            return None
        out = []
        for fname in self.input_files:
            assert(os.path.isfile(fname))
            with open(fname, 'r') as fd:
                icont = fd.read()
            out.append(icont)
        return out

    def check_results(self):
        r"""This should be overridden with checks for the result."""
        if self.output_files is None:
            return
        res_list = self.results
        out_list = self.output_files
        assert(res_list is not None)
        assert(out_list is not None)
        self.assert_equal(len(res_list), len(out_list))
        for res, fout in zip(res_list, out_list):
            self.check_file_exists(fout)
            if isinstance(res, tuple):
                res[0](fout, *res[1:])
            else:
                self.check_file_size(fout, len(res))
                self.check_file_contents(fout, res)

    def run_example(self):
        r"""This runs an example in the correct language."""
        if self.yaml is None:
            if self.name is not None:
                raise unittest.SkipTest("Could not locate example %s in language %s." %
                                        (self.name, self.language))
        else:
            # Copy platform specific makefile
            if self.language == 'make':
                makefile = os.path.join(self.yamldir, 'src', 'Makefile')
                if platform._is_win:  # pragma: windows
                    make_ext = '_windows'
                else:
                    make_ext = '_linux'
                shutil.copy(makefile + make_ext, makefile)
            # Check that language is installed
            for x in self.languages_tested:
                if not tools.is_lang_installed(x):
                    raise unittest.SkipTest("%s not installed." % x)
            # Run
            os.environ.update(self.env)
            self.runner = runner.get_runner(self.yaml, namespace=self.namespace)
            self.runner.run()
            if self.expects_error:
                assert(self.runner.error_flag)
            else:
                assert(not self.runner.error_flag)
            self.check_results()
            self.cleanup()
            # Remove copied makefile
            if self.language == 'make':
                makefile = os.path.join(self.yamldir, 'src', 'Makefile')
                if os.path.isfile(makefile):
                    os.remove(makefile)

    def cleanup(self):
        r"""Cleanup files created during the test."""
        if (self.yaml is not None) and (self.output_files is not None):
            for fout in self.output_files:
                if os.path.isfile(fout):
                    os.remove(fout)

    def test_all(self):
        r"""Test the version of the example that uses all languages."""
        self.language = 'all'
        self.run_example()
        self.language = None

    def test_all_nomatlab(self):
        r"""Test the version of the example that uses all languages."""
        self.language = 'all_nomatlab'
        self.run_example()
        self.language = None

    def test_python(self):
        r"""Test the Python version of the example."""
        self.language = 'python'
        self.run_example()
        self.language = None

    def test_c(self):
        r"""Test the C version of the example."""
        self.language = 'c'
        self.run_example()
        self.language = None

    def test_cpp(self):
        r"""Test the C++ version of the example."""
        self.language = 'cpp'
        self.run_example()
        self.language = None

    def test_make(self):
        r"""Test the Make version of the example."""
        self.language = 'make'
        self.run_example()
        self.language = None

    def test_cmake(self):
        r"""Test the CMake version of the example."""
        self.language = 'cmake'
        self.run_example()
        self.language = None

    def test_matlab(self):  # pragma: matlab
        r"""Test the Matlab version of the example."""
        self.language = 'matlab'
        self.run_example()
        self.language = None



import os
import numpy as np
from yggdrasil import serialize
from yggdrasil.tests import extra_example
from yggdrasil.examples.tests import TestExample


@extra_example
class TestExampleAsciiIO(TestExample):
    r"""Test the AsciiIO example."""

    example_name = 'ascii_io'

    @property
    def input_file(self):
        r"""str: Input file."""
        return os.path.join(self.yamldir, 'Input', 'input_file.txt')

    @property
    def input_table(self):
        r"""str: Input table file."""
        return os.path.join(self.yamldir, 'Input', 'input_table.txt')
        
    @property
    def input_array(self):
        r"""str: Input array file."""
        return os.path.join(self.yamldir, 'Input', 'input_array.txt')

    # @property
    # def input_files(self):
    #     r"""list Input files for the run."""
    #     return [self.input_file, self.input_table, self.input_array]
        
    @property
    def output_file(self):
        r"""str: Output file for the run."""
        for o, yml in self.runner.outputdrivers.items():
            if yml['ocomm_kws']['comm'][0].get('filetype', None) == 'ascii':
                return os.path.join(self.tempdir,
                                    yml['ocomm_kws']['comm'][0]['address'])
        raise Exception('Could not locate output file in yaml.')  # pragma: debug

    @property
    def output_table(self):
        r"""str: Output table for the run."""
        for o, yml in self.runner.outputdrivers.items():
            if (((yml['ocomm_kws']['comm'][0].get('filetype', None) == 'table')
                 and (not yml['ocomm_kws']['comm'][0].get('as_array', False)))):
                return os.path.join(self.tempdir,
                                    yml['ocomm_kws']['comm'][0]['address'])
        raise Exception('Could not locate output table in yaml.')  # pragma: debug

    @property
    def output_array(self):
        r"""str: Output array for the run."""
        for o, yml in self.runner.outputdrivers.items():
            if (((yml['ocomm_kws']['comm'][0].get('filetype', None) == 'table')
                 and (yml['ocomm_kws']['comm'][0].get('as_array', False)))):
                return os.path.join(self.tempdir,
                                    yml['ocomm_kws']['comm'][0]['address'])
        raise Exception('Could not locate output array in yaml.')  # pragma: debug

    @property
    def output_files(self):
        r"""list: Output files for the run."""
        return [self.output_file, self.output_table, self.output_array]

    @property
    def results(self):
        r"""list: Results that should be found in the output files."""
        assert(os.path.isfile(self.input_file))
        assert(os.path.isfile(self.input_table))
        assert(os.path.isfile(self.output_array))
        with open(self.input_file, 'r') as fd:
            icont = fd.read()
        with open(self.input_table, 'rb') as fd:
            iATT = serialize.table_to_array(fd.read(), comment='#')
        with open(self.input_array, 'rb') as fd:
            iATA = serialize.table_to_array(fd.read(), comment='#')
        return [icont,
                (self.check_table, iATT),
                (self.check_table, iATA)]

    def check_table(self, fname, iAT):
        r"""Assert that contents of input/output ascii tables are identical."""
        with open(fname, 'rb') as fd:
            oAT = serialize.table_to_array(fd.read(), comment='#')
        np.testing.assert_equal(oAT, iAT)



from yggdrasil.tests import ErrorClass
from yggdrasil.components import import_component


def ErrorComm(name, base_comm='CommBase', **kwargs):  # pragma: debug
    r"""Wrapper to return errored version of a comm class.

    Args:
        name (str): The environment variable where communication address is
            stored.
        base_comm (str, optional): Name of the base comm that should be used.
            Defaults to 'CommBase'.
        **kwargs: Additional keyword arguments are passed to the class
            constructor.

    Returns:
        ErrorClass: Instance of a comm class that will raise an error at the
            requested locaiton.

    """
    base_class = import_component('comm', base_comm)
    out = ErrorClass(base_class, name, **kwargs)
    return out



import os
import uuid
from yggdrasil.tests import YggTestClassInfo


class TestParam(YggTestClassInfo):
    r"""Test parameters for basic Driver test class.

    Attributes:
        driver (str): Name of driver class.
        args (object): Driver arguments.
        namespace (str): PSI namespace to run drivers in.

    """

    driver = 'Driver'
    args = None

    def __init__(self, *args, **kwargs):
        super(TestParam, self).__init__(*args, **kwargs)
        self.namespace = 'TESTING_%s' % self.uuid
        self.attr_list += ['name', 'sleeptime', 'longsleep', 'timeout',
                           'yml', 'env', 'namespace', 'rank', 'working_dir',
                           'lock']
        self._inst_kwargs = {'yml': {'working_dir': self.working_dir},
                             'timeout': self.timeout,
                             'sleeptime': self.sleeptime,
                             # 'working_dir': self.working_dir,
                             'namespace': self.namespace}
        self.debug_flag = False

    def get_fresh_name(self):
        r"""Get a fresh name for a new instance that won't overlap with the base."""
        return 'Test%s_%s' % (self.cls, str(uuid.uuid4()))
    
    @property
    def working_dir(self):
        r"""str: Working directory."""
        return os.path.dirname(__file__)

    @property
    def skip_start(self):
        r"""bool: True if driver shouldn't be started. False otherwise."""
        return ('NoStart' in str(self.__class__))

    @property
    def skip_init(self):
        r"""bool: True fi driver shouldn't be initialized during startup.
        False otherwise."""
        return ('NoInit' in str(self.__class__))

    @property
    def cls(self):
        r"""str: Driver class."""
        return self.driver

    @property
    def mod(self):
        r"""str: Absolute path to module containing driver."""
        return 'yggdrasil.drivers.%s' % self.cls

    @property
    def inst_args(self):
        r"""tuple: Driver arguments."""
        out = [self.name]
        if self.args is not None:
            out.append(self.args)
        return out

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        out = super(TestParam, self).inst_kwargs
        out['timeout'] = self.timeout
        out['sleeptime'] = self.sleeptime
        return out

    def setup(self, *args, **kwargs):
        r"""Create a driver instance and start the driver."""
        super(TestParam, self).setup(*args, **kwargs)
        if not (self.skip_init or self.skip_start):
            self.instance.start()

    @property
    def name(self):
        r"""str: Name of the test driver."""
        return 'Test%s_%s' % (self.cls, self.uuid)

    def create_instance(self, *args, **kwargs):
        r"""Create a new instance object."""
        curpath = os.getcwd()
        try:
            os.chdir(self.working_dir)
            inst = super(TestParam, self).create_instance(*args, **kwargs)
        finally:
            os.chdir(curpath)
        return inst

    def remove_instance(self, inst):
        r"""Remove an instance."""
        if not inst.was_terminated:
            inst.terminate()
        if inst.is_alive():  # pragma: debug
            inst.join()
        inst.cleanup()
        assert(not inst.is_alive())
        super(TestParam, self).remove_instance(inst)


class TestDriverNoInit(TestParam):
    r"""Test runner for driver without initializing the driver."""
    pass

        
class TestDriverNoStart(TestParam):
    r"""Test runner for basic Driver class without starting driver."""

    def setup(self, *args, **kwargs):
        r"""Create a driver instance without starting the driver."""
        super(TestDriverNoStart, self).setup(*args, **kwargs)
        assert(not self.instance.is_alive())

    def test_attributes(self):
        r"""Assert that the driver has all of the required attributes."""
        for a in self.attr_list:
            if not hasattr(self.instance, a):  # pragma: debug
                raise AttributeError("Driver does not have attribute %s" % a)


class TestDriver(TestParam):
    r"""Test runner for basic Driver class."""

    def assert_before_stop(self):
        r"""Assertions to make before stopping the driver instance."""
        pass

    def run_before_terminate(self):
        r"""Commands to run while the instance is running, before terminate."""
        pass

    def run_before_stop(self):
        r"""Commands to run while the instance is running."""
        pass

    def assert_after_terminate(self):
        r"""Assertions to make after terminating the driver instance."""
        assert(not self.instance.is_alive())

    def assert_after_stop(self):
        r"""Assertions to make after stopping the driver instance."""
        self.assert_after_terminate()

    def test_init_del(self):
        r"""Test driver creation and deletion."""
        self.instance.printStatus()

    def test_run_stop(self):
        r"""Start the thread, then stop it."""
        self.assert_before_stop()
        self.run_before_stop()
        self.instance.wait(self.sleeptime)
        self.instance.stop()
        self.instance.stop()
        self.assert_after_stop()

    def test_run_terminate(self):
        r"""Start the thread, then terminate it."""
        self.assert_before_stop()
        self.run_before_terminate()
        self.instance.terminate()
        # Second time to ensure it is escaped
        self.instance.terminate()
        if self.instance.is_alive():  # pragma: debug
            self.instance.join()
        self.assert_after_terminate()



import copy
from yggdrasil.metaschema.datatypes import encode_type, compare_schema
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


class ItemsMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'items' property."""

    name = 'items'
    _replaces_existing = True
    _validate = False

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'items' container property."""
        if isinstance(typedef, (list, tuple)):
            typedef_list = typedef
        else:
            typedef_list = [copy.deepcopy(typedef) for x in instance]
        assert(len(typedef_list) == len(instance))
        return [encode_type(v, typedef=t) for v, t in zip(instance, typedef_list)]

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparison method for 'items' container property."""
        if isinstance(prop1, dict) and isinstance(prop2, dict):
            for e in compare_schema(prop1, prop2, root1=root1, root2=root2):
                yield e
            return
        elif isinstance(prop1, dict) and isinstance(prop2, cls.python_types):
            for p2 in prop2:
                for e in compare_schema(prop1, p2, root1=root1, root2=root2):
                    yield e
            return
        elif isinstance(prop1, cls.python_types) and isinstance(prop2, dict):
            for p1 in prop1:
                for e in compare_schema(p1, prop2, root1=root1, root2=root2):
                    yield e
            return
        elif not (isinstance(prop1, cls.python_types)
                  and isinstance(prop2, cls.python_types)):
            yield "Values have incorrect type: %s, %s." % (type(prop1), type(prop2))
            return
        if len(prop1) != len(prop2):
            yield 'Unequal number of elements. %d vs. %d' % (len(prop1), len(prop2))
        for p1, p2 in zip(prop1, prop2):
            for e in compare_schema(p1, p2, root1=root1, root2=root2):
                yield e



from yggdrasil.components import import_component
from yggdrasil.communication import CommBase, get_comm


class ServerComm(CommBase.CommBase):
    r"""Class for handling Server side communication.

    Args:
        name (str): The environment variable where communication address is
            stored.
        request_comm (str, optional): Comm class that should be used for the
            request comm. Defaults to None.
        response_kwargs (dict, optional): Keyword arguments for the response
            comm. Defaults to empty dict.
        **kwargs: Additional keywords arguments are passed to the input comm.

    Attributes:
        response_kwargs (dict): Keyword arguments for the response comm.
        icomm (Comm): Request comm.
        ocomm (Comm): Response comm for last request.

    """

    _dont_register = True
    
    def __init__(self, name, request_comm=None, response_kwargs=None,
                 dont_open=False, **kwargs):
        if response_kwargs is None:
            response_kwargs = dict()
        icomm_name = name
        icomm_kwargs = kwargs
        icomm_kwargs['direction'] = 'recv'
        icomm_kwargs['dont_open'] = True
        icomm_kwargs['comm'] = request_comm
        self.response_kwargs = response_kwargs
        self.icomm = get_comm(icomm_name, **icomm_kwargs)
        self.ocomm = None
        self.response_kwargs.setdefault('comm', self.icomm.comm_class)
        self.response_kwargs.setdefault('recv_timeout', self.icomm.recv_timeout)
        self._used_response_comms = dict()
        super(ServerComm, self).__init__(self.icomm.name, dont_open=dont_open,
                                         recv_timeout=self.icomm.recv_timeout,
                                         is_interface=self.icomm.is_interface,
                                         direction='recv', no_suffix=True,
                                         address=self.icomm.address)

    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked.

        Returns:
            bool: Is the comm installed.

        """
        return import_component('comm').is_installed(language=language)

    @property
    def maxMsgSize(self):
        r"""int: Maximum size of a single message that should be sent."""
        return self.icomm.maxMsgSize
        
    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return import_component('comm').underlying_comm_class()

    @classmethod
    def comm_count(cls):
        r"""int: Number of communication connections."""
        return import_component('comm').comm_count()

    @classmethod
    def new_comm_kwargs(cls, name, request_comm=None, **kwargs):
        r"""Initialize communication with new comms.

        Args:
            name (str): Name for new comm.
            request_comm (str, optional): Name of class for new input comm.
                Defaults to None.

        """
        args = [name]
        icomm_class = import_component('comm', request_comm)
        kwargs['direction'] = 'recv'
        if 'address' not in kwargs:
            iargs, kwargs = icomm_class.new_comm_kwargs(name, **kwargs)
        kwargs['request_comm'] = request_comm
        return args, kwargs

    @property
    def opp_comms(self):
        r"""dict: Name/address pairs for opposite comms."""
        out = super(ServerComm, self).opp_comms
        out.update(**self.icomm.opp_comms)
        return out

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(ServerComm, self).opp_comm_kwargs()
        kwargs['comm'] = "ClientComm"
        kwargs['request_comm'] = self.icomm.comm_class
        kwargs['response_kwargs'] = self.response_kwargs
        return kwargs
        
    def open(self):
        r"""Open the connection."""
        super(ServerComm, self).open()
        self.icomm.open()

    def close(self, *args, **kwargs):
        r"""Close the connection."""
        self.icomm.close(*args, **kwargs)
        if self.ocomm is not None:
            self.ocomm.close()
        for ocomm in self._used_response_comms.values():
            ocomm.close()
        super(ServerComm, self).close(*args, **kwargs)

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return self.icomm.is_open

    @property
    def is_closed(self):
        r"""bool: True if the connection is closed."""
        return self.icomm.is_closed

    @property
    def n_msg_recv(self):
        r"""int: The number of messages in the connection."""
        return self.icomm.n_msg_recv

    @property
    def n_msg_recv_drain(self):
        r"""int: The number of messages in the connection to drain."""
        return self.icomm.n_msg_recv_drain

    # RESPONSE COMM
    def create_response_comm(self):
        r"""Create a response comm based on information from the last header."""
        if not isinstance(self.icomm._last_header, dict):  # pragma: debug
            raise RuntimeError("No header received with last message.")
        elif 'response_address' not in self.icomm._last_header:  # pragma: debug
            raise RuntimeError("Last header does not contain response address.")
        comm_kwargs = dict(address=self.icomm._last_header['response_address'],
                           direction='send', is_response_server=True,
                           single_use=True, **self.response_kwargs)
        self.ocomm = get_comm(self.name + '.server_response_comm',
                              **comm_kwargs)

    def remove_response_comm(self):
        r"""Remove response comm."""
        self.icomm._last_header = None
        # self.ocomm.close_on_empty(no_wait=True)
        self._used_response_comms[self.ocomm.name] = self.ocomm
        self.ocomm = None

    # SEND METHODS
    def send(self, *args, **kwargs):
        r"""Send a message to the output comm.

        Args:
            *args: Arguments are passed to output comm send method.
            **kwargs: Keyword arguments are passed to output comm send method.

        Returns:
            obj: Output from output comm send method.

        """
        # if self.is_closed:
        #     self.debug("send(): Connection closed.")
        #     return False
        if self.ocomm is None:  # pragma: debug
            raise RuntimeError("There is no registered response comm.")
        out = self.ocomm.send(*args, **kwargs)
        self.remove_response_comm()
        return out

    # RECV METHODS
    def recv(self, *args, **kwargs):
        r"""Receive a message from the input comm and open a new response comm
        for output using address from the header.

        Args:
            *args: Arguments are passed to input comm recv method.
            **kwargs: Keyword arguments are passed to input comm recv method.

        Returns:
            obj: Output from input comm recv method.

        """
        # if self.is_closed:
        #     self.debug("recv(): Connection closed.")
        #     return (False, None)
        flag, msg = self.icomm.recv(*args, **kwargs)
        if flag and msg and (msg != self.eof_msg):
            self.create_response_comm()
        return flag, msg

    # OLD STYLE ALIASES
    def rpcSend(self, *args, **kwargs):
        r"""Alias for RPCComm.send"""
        return self.send(*args, **kwargs)

    def rpcRecv(self, *args, **kwargs):
        r"""Alias for RPCComm.recv"""
        return self.recv(*args, **kwargs)
    
    def drain_messages(self, direction='recv', **kwargs):
        r"""Sleep while waiting for messages to be drained."""
        if direction == 'recv':
            self.icomm.drain_messages(direction='recv', **kwargs)

    def purge(self):
        r"""Purge input and output comms."""
        self.icomm.purge()
        # Not sure if server should purge the response queue...
        # if self.ocomm is not None:
        #     self.ocomm.purge()
        super(ServerComm, self).purge()



import os
from yggdrasil.examples import source, yamls


rst_dir = os.path.dirname(os.path.abspath(__file__))
toc_file = os.path.join(rst_dir, 'examples_toc.rst')
lang2print = {'python': 'Python',
              'matlab': 'Matlab',
              'cmake': 'CMake',
              'make': 'Make',
              'r': 'R',
              'R': 'R',
              'c': 'C',
              'cpp': 'C++',
              'all': 'Mixed',
              'all_nomatlab': 'Mixed w/o Matlab'}
_default_lang = 'python'


def get_file(fname, local=False):
    if local:
        return os.path.join(rst_dir, fname)
    else:
        return fname


def get_rst_file(k, local=False):
    return get_file('%s.rst' % (k), local=local)


def get_html_file(k, local=False):
    return get_file('%s.html' % (k), local=local)


def get_src_file(k, local=False):
    return get_file('%s_src.rst' % (k), local=local)


def get_yml_file(k, local=False):
    return get_file('%s_yml.rst' % (k), local=local)


def make_toc_file(key_list):
    with open(toc_file, 'w') as fd:
        write_toc_file(fd, key_list)


def make_rst_file(k):
    fname = get_rst_file(k, local=True)
    with open(fname, 'w') as fd:
        write_rst(fd, k)
    make_src_file(k)
    make_yml_file(k)

        
def make_src_file(k):
    fname = get_src_file(k, local=True)
    with open(fname, 'w') as fd:
        write_src_ref(fd, k)

        
def make_yml_file(k):
    fname = get_yml_file(k, local=True)
    with open(fname, 'w') as fd:
        write_yml_ref(fd, k)

        
def get_rel_path(fname, upone=False):
    if upone:
        top_dir = os.path.dirname(rst_dir)
    else:
        top_dir = rst_dir
    return os.path.relpath(fname, top_dir)


def write_src_ref(fd, k):
    write_src(fd, k, _default_lang, upone=True)
    fd.write('\n')
    write_ref_link(fd, k)

    
def write_yml_ref(fd, k):
    write_yml(fd, k, _default_lang, upone=True)
    fd.write('\n')
    write_ref_link(fd, k)


def write_ref_link(fd, k):
    fd.write("(`%s <%s>`__)\n" % ('Example in other languages',
                                  os.path.join('examples', get_html_file(k))))


def write_toc_file(fd, key_list):
    head = "Examples"
    fd.write(head + '\n')
    fd.write(len(head)*'=' + '\n\n')
    fd.write(".. toctree::\n\n")
    for k in key_list:
        fd.write("   %s\n" % get_rst_file(k).split('.rst')[0])
    fd.write("\n")
        

def write_rst(fd, k):
    head = '%s Example' % k
    fd.write(head + '\n')
    fd.write(len(head)*'=' + '\n\n')
    for l in source[k]:
        write_lang(fd, k, l)
        fd.write('\n')


def write_lang(fd, k, l):
    head = '%s Version' % lang2print[l]
    fd.write(head + '\n')
    fd.write(len(head)*'-' + '\n\n')
    write_src(fd, k, l)
    fd.write('\n')
    write_yml(fd, k, l)
    fd.write('\n')

    
def write_code_line(fd, s, upone=False, language=None):
    p = os.path.sep + get_rel_path(s, upone=True)
    ext2lang = {'.yml': 'yaml', '.py': 'python',
                '.c': 'c', '.cpp': 'c++', '.m': 'matlab'}
    if language is None:
        ext = os.path.splitext(p)[-1]
        language = ext2lang.get(ext, 'python')
    fd.write(".. literalinclude:: %s\n" % p)
    fd.write("   :language: %s\n" % language)
    fd.write("   :linenos:\n")
    # fd.write(".. include:: %s\n" % get_rel_path(s, upone=upone))
    # fd.write("   :code: %s\n" % language)
    # fd.write("   :number-lines:\n")
    fd.write("\n")


def write_src(fd, k, l, upone=False):
    fd.write("Model Code:\n\n")
    if isinstance(source[k][l], list):
        for s in source[k][l]:
            write_code_line(fd, s, upone=upone)
    else:
        write_code_line(fd, source[k][l], upone=upone)

    
def write_yml(fd, k, l, upone=False):
    fd.write("Model YAML:\n\n")
    if isinstance(yamls[k][l], list):
        for y in yamls[k][l]:
            write_code_line(fd, y, upone=upone, language='yaml')
    else:
        write_code_line(fd, yamls[k][l], upone=upone, language='yaml')

        
# rst_examples = source.keys()  # all examples
rst_examples = ['gs_lesson%d' % x for x in range(1, 5)]
rst_examples.append('gs_lesson4b')  # Special case
rst_examples += ['formatted_io%d' % x for x in range(1, 10)]
rst_examples += ['rpc_lesson%d' % x for x in range(1, 3)]
make_toc_file(rst_examples)
for k in rst_examples:
    make_rst_file(k)



from __future__ import print_function
from yggdrasil.interface.YggInterface import YggInput, YggOutput


def run():
    print('Hello from Python pipe_dst')

    # Ins/outs matching with the the model yaml
    inq = YggInput('input_pipe')
    outf = YggOutput('output_file')
    print("pipe_dst(P): Created I/O channels")

    # Continue receiving input from the queue
    count = 0
    while True:
        ret, buf = inq.recv()
        if not ret:
            print("pipe_dst(P): Input channel closed")
            break
        ret = outf.send(buf)
        if not ret:
            raise RuntimeError("pipe_dst(P): SEND ERROR ON MSG %d" % count)
        count += 1

    print('Goodbye from Python destination. Received %d messages.' % count)

    
if __name__ == '__main__':
    run()



from yggdrasil.metaschema.datatypes import (
    get_registered_types, get_type_class, MetaschemaTypeError)
from yggdrasil.metaschema.properties.MetaschemaProperty import MetaschemaProperty


def _specificity_sort_key(item):
    return -item[1].specificity


class TypeMetaschemaProperty(MetaschemaProperty):
    r"""Type property with validation of new properties."""

    name = 'type'
    _replaces_existing = True
    _validate = False

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Method to encode the property given the object.

        Args:
            instance (object): Object to get property for.
            typedef (object, None): Template value in type definition to use
                for initializing encoding some cases. Defaults to None and
                is ignored.

        Returns:
            object: Encoded property for instance.

        """
        type_registry = get_registered_types()
        for t, cls in sorted(type_registry.items(), key=_specificity_sort_key):
            if (t != 'any') and cls.validate(instance):
                return t
        raise MetaschemaTypeError(
            "Could not encode 'type' property for Python type: %s"
            % type(instance))

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Method to determine compatiblity of one property value with another.
        This method is not necessarily symmetric in that the second value may
        not be compatible with the first even if the first is compatible with
        the second.

        Args:
            prop1 (object): Property value to compare against prop2.
            prop2 (object): Property value to compare against.
            
        Yields:
            str: Comparision failure messages.

        """
        type_cls = get_type_class(prop1)
        if not type_cls.issubtype(prop2):
            yield "Type '%s' is not a subtype of type '%s'" % (prop1, prop2)

    @classmethod
    def normalize(cls, normalizer, value, instance, schema):
        r"""Method to normalize the instance based on the property value.

        Args:
            normalizer (Normalizer): Normalizer class.
            value (object): Property value.
            instance (object): Object to normalize.
            schema (dict): Schema containing this property.

        Returns:
            object: Normalized object.

        """
        if isinstance(value, (list, tuple)):
            v0 = value[0]
            for v in value:
                t = get_type_class(v)
                # if normalizer.is_type(instance, v):
                if t.validate(v):
                    v0 = v
                    break
            type_cls = get_type_class(v0)
        else:
            type_cls = get_type_class(value)
        return type_cls.normalize(instance)



import re
import os
import textwrap


def write_table(lines, fname=None, fname_base=None, fname_dir=None,
                verbose=False, **kwargs):
    r"""Write a table to a file.

    Args:
        fname (str, optional): Name of the file where the table should be written.
            If not provided, one is constructed from fname_base and fname_dir.
        fname_base (str, optional): Base name of the file where the table should
            be written if fname is not provided. If fname is None, fname_base
            must be provided.
        fname_dir (str, optional): Directory where the table should be written
            if fname is not provided. If not provided, this defaults to the
            current working directory.
        verbose (bool, optional): If True, the lines are printed to standard
            out in addition to the file. Defaults to False.
        **kwargs: Additional keyword arguments are ignored.

    Returns:
        str: Name of file created.

    Raises:
        ValueError: If neither fname nor fname_base are provided.

    """
    if fname is None:
        if fname_base is None:
            raise ValueError("If fname is not provided, fname_base must be.")
        if fname_dir is None:
            fname_dir = os.getcwd()
        fname = os.path.join(fname_dir, fname_base)
    else:
        fname_dir, fname_base = os.path.split(fname)
    fname_ref = fname_base.replace('.rst', '_rst')
    # Write
    lines = ['.. _%s:' % fname_ref, ''] + lines
    with open(fname, 'w') as fd:
        fd.write('\n'.join(lines))
    if verbose:
        print('\n'.join(lines))
    return fname


def write_comm_devnotes_table(**kwargs):
    r"""Write a table containing development notes for the registered comms.

    Args:
        **kwargs: Additional keyword arguments are passed to dict2table and
            write_table.

    Returns:
        str, list: Name of file or files created.

    """
    from yggdrasil import tools, components
    kwargs.setdefault('fname_base', 'comm_devnotes_table.rst')
    args = {}
    for comm in tools.get_supported_comm():
        if comm in ['default', 'rmq_async']:
            continue
        cdrv = components.import_component('comm', comm)
        args[comm] = {'developer notes': get_docs_section(cdrv.__doc__,
                                                          keys=['Developer Notes:',
                                                                'Development Notes:'],
                                                          join_lines=True)}
        if cdrv.address_description is not None:
            args[comm]['address'] = cdrv.address_description
    kwargs.setdefault('key_column_name', 'comm')
    kwargs.setdefault('val_column_name', 'developer notes')
    kwargs.setdefault('wrapped_columns', {'address': 40,
                                          'developer notes': 80})
    kwargs.setdefault('column_order', ['comm', 'address', 'developer notes'])
    lines = dict2table(args, **kwargs)
    return write_table(lines, **kwargs)


def write_datatype_mapping_table(**kwargs):
    r"""Write a table containing mapping of datatypes between different
    languages.

    Args:
        **kwargs: Additional keyword arguments are passed to dict2table and
            write_table.

    Returns:
        str, list: Name of file or files created.

    """
    from yggdrasil import tools, components
    from yggdrasil.metaschema.datatypes import _type_registry
    kwargs.setdefault('fname_base', 'datatype_mapping_table.rst')
    args = {}
    for k, v in _type_registry.items():
        if v.cross_language_support:
            args[k] = {'notes': get_docs_section(v.__doc__,
                                                 keys=['Developer Notes:',
                                                       'Development Notes:'],
                                                 join_lines=True)}
    for lang in tools.get_supported_lang():
        if lang in ['lpy', 'make', 'cmake', 'executable']:
            continue
        if lang == 'cpp':
            lang = 'c++'
        ldrv = components.import_component('model', lang)
        for k in args.keys():
            entry = ldrv.type_map.get(k, '')
            if entry:
                args[k][lang] = '``%s``' % entry
            else:
                args[k][lang] = ''
    kwargs.setdefault('key_column_name', 'schema')
    kwargs.setdefault('val_column_name', 'notes')
    kwargs.setdefault('prune_empty_columns', False)
    kwargs.setdefault('last_column', 'notes')
    # kwargs.setdefault('column_order', ['schema', 'notes'])
    kwargs.setdefault('wrapped_columns', {'notes': 40})
    lines = dict2table(args, **kwargs)
    return write_table(lines, **kwargs)
    

def write_datatype_table(table_type='all', **kwargs):
    r"""Write a table containing entries from the descriptions of datatypes.

    Args:
        table_type (str, optional): Type of table that should be created for the
            class. Defaults to 'all'. Supported values include:

            * 'all': Create each type of table.
            * 'simple': Create a table of standard JSON datatypes.
            * 'container': Create a table of container datatypes.
            * 'yggdrasil': Create a table of yggdrasil specific datatypes.

        **kwargs: Additional keyword arguments are passed to dict2table and
            write_table.

    Returns:
        str, list: Name of file or files created.

    Raises:
        ValueError: If table_type is not one of the supported values.

    """
    from yggdrasil.metaschema.datatypes import _type_registry
    table_type_list = ['simple', 'container', 'yggdrasil']
    if table_type == 'all':
        fname = kwargs.get("fname", None)
        fname_base = kwargs.get("fname_base", None)
        assert((fname is None) and (fname_base is None))
        out = []
        for k in table_type_list:
            out.append(write_datatype_table(table_type=k, **kwargs))
        return out
    elif table_type not in table_type_list:
        raise ValueError("Unsupported table_type: '%s'" % table_type)
    fname_format = 'datatype_table_%s.rst'
    kwargs.setdefault('fname_base', fname_format % (table_type))
    target_types = []
    args = {}
    if table_type == 'simple':
        for k, v in _type_registry.items():
            if v._replaces_existing and (not hasattr(v, '_container_type')):
                target_types.append(k)
    elif table_type == 'container':
        for k, v in _type_registry.items():
            if v._replaces_existing and hasattr(v, '_container_type'):
                target_types.append(k)
    elif table_type == 'yggdrasil':
        for k, v in _type_registry.items():
            if not v._replaces_existing:
                target_types.append(k)
    for k in target_types:
        v = _type_registry[k]
        args[k] = {'description': v.description,
                   'required properties': v.definition_properties}
    kwargs.setdefault('key_column_name', 'type')
    kwargs.setdefault('list_columns', 'required properties')
    lines = dict2table(args, **kwargs)
    return write_table(lines, **kwargs)
    

def write_classdocs_table(class_, table_type='all', **kwargs):
    r"""Write a table containing entries from a section in a class's
    documentation.

    Args:
        class_ (type): Class with docstring that should be parsed.
        table_type (str, optional): Type of table that should be created for the
            class. Defaults to 'all'. Supported values include:

            * 'all': Create each type of table for the specified class.
            * 'args': Create a table of the class's argumetns.
            * 'attr': Create a table of the class's attributes.
            * 'classattr': Create a table of the class's class attributes.

        **kwargs: Additional keyword arguments are passed to dict2table and
            write_table.

    Returns:
        str, list: Name of file or files created.

    Raises:
        ValueError: If table_type is not one of the supported values.

    """
    table_type2keys = {'args': ['Args:', 'Arguments:'],
                       'attr': ['Attr:', 'Attributes:'],
                       'classattr': ['Class Attr:', 'Class Attributes']}
    table_type_list = list(table_type2keys.keys())
    if table_type == 'all':
        fname = kwargs.get("fname", None)
        fname_base = kwargs.get("fname_base", None)
        assert((fname is None) and (fname_base is None))
        out = []
        for k in table_type_list:
            out.append(write_classdocs_table(class_, table_type=k, **kwargs))
        return out
    elif table_type not in table_type2keys:
        raise ValueError("Unsupported table_type: '%s'" % table_type)
    fname_format = 'class_table_%s_%s.rst'
    kwargs.setdefault('fname_base', fname_format % (class_.__name__, table_type))
    args = docs2args(class_.__doc__, keys=table_type2keys[table_type])
    if table_type == 'args':
        for k, v in args.items():
            if v.get('type', '').endswith(', optional'):
                v['type'] = v['type'].split(', optional')[0]
                v['required'] = ''
            else:
                v['required'] = 'X'
    else:
        for k, v in args.items():
            if v.get('description', '').endswith(' [REQUIRED]'):
                v['description'] = v['description'].split(' [REQUIRED]')[0]
                v['required'] = 'X'
            else:
                v['required'] = ''
    lines = dict2table(args, **kwargs)
    return write_table(lines, **kwargs)


def write_component_table(comp='all', table_type='all', **kwargs):
    r"""Write a component table to a file.

    Args:
        comp (str): Name of a component type to create a table for. Defaults to
            'all' and tables are created for each of the registered components.
        table_type (str, optional): Type of table that should be created for the
            component. Defaults to 'all'. Supported values include:

            * 'all': Create each type of table for the specified component.
            * 'general': Create a table of properties common to all components
              of the specified type.
            * 'specific': Create a table of properties specific to only some
              components of the specified type.
            * 'subtype': Create a table describing the subtypes available for
              the specified component type.

        **kwargs: Additional keyword arguments are passed to component2table
            and write_table.

    Returns:
        str, list: Name of file or files created.

    """
    from yggdrasil.schema import get_schema
    s = get_schema()
    table_type_list = ['subtype', 'general', 'specific']
    # Loop
    if comp == 'all':
        fname = kwargs.get("fname", None)
        fname_base = kwargs.get("fname_base", None)
        assert((fname is None) and (fname_base is None))
        out = []
        for k in s.keys():
            new_out = write_component_table(comp=k, table_type=table_type, **kwargs)
            if isinstance(new_out, list):
                out += new_out
            else:
                out.append(new_out)
        return out
    if table_type == 'all':
        fname = kwargs.get("fname", None)
        fname_base = kwargs.get("fname_base", None)
        assert((fname is None) and (fname_base is None))
        out = []
        for k in table_type_list:
            out.append(write_component_table(comp=comp, table_type=k, **kwargs))
        return out
    # Construct file name
    fname_format = 'schema_table_%s_%s.rst'
    kwargs.setdefault('subtype_ref',
                      (fname_format % (comp, 'subtype')).replace('.rst', '_rst'))
    kwargs.setdefault('fname_base', fname_format % (comp, table_type))
    lines = component2table(comp, table_type, **kwargs)
    return write_table(lines, **kwargs)


def component2table(comp, table_type, include_required=None,
                    subtype_ref=None, **kwargs):
    r"""Create a table describing a component.

    Args:
        comp (str): Name of a component type to create a table for.
        table_type (str): Type of table that should be created for the component.
            Supported values include:

            * 'general': Create a table of properties common to all components
              of the specified type.
            * 'specific': Create a table of properties specific to only some
              components of the specified type.
            * 'subtype': Create a table describing the subtypes available for
              the specified component type.

        include_required (bool, optional): If True, a required column is
            included. Defaults to True if table_type is 'general' and False
            otherwise.
        subtype_ref (str, optional): Reference for the subtype table for the
            specified component that should be used in the description for the
            subtype property. Defaults to None and is ignored.
        **kwargs: Additional keyword arguments are passed to dict2table.

    Returns:
        list: Lines comprising the table.

    """
    from yggdrasil.schema import get_schema
    if include_required is None:
        if table_type == 'general':
            include_required = True
        else:
            include_required = False
    if table_type in ['subtype', 'specific']:
        kwargs.setdefault('prune_empty_columns', True)
    args = {}
    s = get_schema()
    subtype_key = s[comp].subtype_key
    if table_type in ['general', 'specific']:
        # Set defaults
        kwargs.setdefault('key_column_name', 'option')
        kwargs.setdefault('val_column_name', 'description')
        kwargs.setdefault('column_order', [kwargs['key_column_name'],
                                           'type', 'required',
                                           kwargs['val_column_name']])
        if (not include_required) and ('required' in kwargs['column_order']):
            kwargs['column_order'].remove('required')
        # Get list of component subtypes
        if table_type == 'general':
            s_comp_list = [s[comp].get_subtype_schema('base', unique=True)]
        else:
            s_comp_list = [s[comp].get_subtype_schema(x, unique=True)
                           for x in s[comp].classes]
        # Loop over subtyeps
        out_apply = {}
        for s_comp in s_comp_list:
            for k, v in s_comp['properties'].items():
                if (k == subtype_key) and (table_type == 'specific'):
                    continue
                if k not in args:
                    args[k] = {'type': v.get('type', ''),
                               'description': v.get('description', '')}
                    if include_required:
                        if k in s_comp.get('required', []):
                            args[k]['required'] = ''
                        else:
                            args[k]['required'] = 'X'
                if (table_type == 'specific'):
                    if k not in out_apply:
                        out_apply[k] = []
                    out_apply[k] += s_comp['properties'][subtype_key]['enum']
                elif (subtype_ref is not None) and (k == subtype_key):
                    args[k]['description'] += (
                        ' (Options described :ref:`here <%s>`)' % subtype_ref)
        if table_type == 'specific':
            for k, v in args.items():
                v['Valid for \'%s\' of' % subtype_key] = list(set(out_apply[k]))
    elif table_type == 'subtype':
        kwargs.setdefault('key_column_name', subtype_key)
        for x, subtypes in s[comp].schema_subtypes.items():
            s_comp = s[comp].get_subtype_schema(x, unique=True)
            subt = subtypes[0]
            args[subt] = {
                'description': s_comp['properties'][subtype_key].get(
                    'description', '')}
            if len(subtypes) > 1:
                args[subt]['aliases'] = subtypes[1:]
            if s_comp['properties'][subtype_key].get('default', None) in subtypes:
                args[subt]['description'] = ('[DEFAULT] '
                                             + args[subt]['description'])
    else:
        raise ValueError("Unsupported table_type: '%s'" % table_type)
    return dict2table(args, **kwargs)


def dict2table(args, key_column_name='option', val_column_name='description',
               column_order=None, wrapped_columns=None, list_columns=None,
               sort_on_key=True, last_column=None, prune_empty_columns=False,
               style='simple', **kwargs):
    r"""Convert a dictionary to a table.

    Args:
        args (dict): Dictionary of entries for the table. The keys will go in
            the first column (the key column) as named by key_column_name.
            If the values are dictionaries, the elements in each dictionary
            will be put in different columns. If the values are strings, they
            will be put in the column identified by val_column_name.
        key_column_name (str, optional): Label that the key column should have.
            Defaults to 'option'.
        val_column_name (str, optional): Lable that the value column should have.
            Defaults to 'description'.
        column_order (list, optional): List specifying the roder that fields
            should be added to the table as columns. Defaults to None and is set
            to [key_column_name, val_column_name]. If a column is missing from
            an entry in args, an empty value will be added to the table in its
            place. Columns in args entries that are not in column_order are
            appended to column_order in sorted order.
        wrapped_columns (dict, optional): Dictionary specifying fields that
            should be wrapped as columns and the widths that the corresponding
            column should be wrapped to. Defaults to {'Description': 80}.
        list_columns (list, optional): List of fields with values that should
            be formatted with one item per line if it is a list. Defaults to
            None and is ignored.
        sort_on_key (bool, optional): If True, the entries in args are added
            as rows in the order determined by sorting on the keys. If False,
            the order will be determine by prop (which is not deterministic
            if a Python 2 dictionary). Defaults to True.
        last_column (str, optional): Name of the field that should be placed
            in the last column position. Defaults to None and is ignored.
        prune_empty_columns (bool, optional): If True, empty columns will be
            removed. If False, they will be included. Defaults to False.
        style (str, optional): Style of table that should be used. 'simple'
            dosn't allow for multi-column or row cells, while 'complex'
            does. Defaults to 'simple'.
        **kwargs: Additional keyword arguments are ignored.
            
    Returns:
        list: Lines comprising the table.

    """
    if wrapped_columns is None:
        wrapped_columns = {'description': 80}
    if list_columns is None:
        list_columns = []
    else:
        style = 'complex'
    # Determine column order
    if column_order is None:
        column_order = [key_column_name, val_column_name]
    # Create dictionary of columns
    columns = {k: [] for k in column_order}
    pos = 0
    prop_order = list(args.keys())
    if sort_on_key:
        prop_order = sorted(prop_order)
    for k in prop_order:
        v = args[k]
        if not isinstance(v, dict):
            v = {val_column_name: v}
        for pk in v.keys():
            if pk not in columns:
                columns[pk] = pos * ['']
        for pk in columns.keys():
            if pk == key_column_name:
                columns[key_column_name].append(k)
            elif pk in list_columns:
                x = v.get(pk, [])
                if not isinstance(x, list):
                    x = [x]
                columns[pk].append(x)
            else:
                columns[pk].append(str(v.get(pk, '')))
        pos += 1
    # Add non-standard fields
    for k in sorted(columns.keys()):
        if k not in column_order:
            column_order.append(k)
    if (last_column is not None) and (last_column in column_order):
        column_order.remove(last_column)
        column_order.append(last_column)
    # Prune empty columns
    column_widths = {}
    for k in list(columns.keys()):
        if len(columns[k]) == 0:
            column_widths[k] = 0
        else:
            column_widths[k] = len(max(columns[k], key=len))
        if prune_empty_columns and (column_widths[k] == 0):
            del column_widths[k]
            del columns[k]
            column_order.remove(k)
    # Get sizes that include headers and wrapping
    for k in columns.keys():
        if k in wrapped_columns:
            w = wrapped_columns[k]
        elif k in list_columns:
            w = max([len(max(irow, key=len)) for irow in columns[k]])
        else:
            w = column_widths[k]
        column_widths[k] = max(w, len(k))
    # Create format string
    if style == 'simple':
        column_beg = ''
        column_end = ''
        column_sep = '   '
        hline_beg = ''
        hline_end = ''
        hline_char = '='
        hline_sep = column_sep
    else:
        column_beg = '| '
        column_end = ' |'
        column_sep = ' | '
        hline_beg = '+-'
        hline_end = '-+'
        hline_char = '-'
        hline_sep = '-+-'
    column_format = (column_beg
                     + column_sep.join(['%-' + str(column_widths[k]) + 's'
                                        for k in column_order])
                     + column_end)
    divider = (hline_beg
               + hline_sep.join([hline_char * column_widths[k]
                                 for k in column_order])
               + hline_end)
    header = column_format % tuple([k.title() for k in column_order])
    # Table
    if len(columns) == 0:
        pos = 0
    else:
        pos = len(columns[list(columns.keys())[0]])
    lines = [divider, header, divider]
    for i in range(pos):
        row = []
        max_row_len = 1
        for k in column_order:
            if k in wrapped_columns:
                row.append(textwrap.wrap(columns[k][i], wrapped_columns[k]))
            elif k in list_columns:
                row.append(columns[k][i])
            else:
                row.append([columns[k][i]])
            max_row_len = max(max_row_len, len(row[-1]))
        for j in range(len(row)):
            row[j] += (max_row_len - len(row[j])) * ['']
        for k in range(max_row_len):
            lines.append(column_format % tuple([row[j][k] for j in range(len(row))]))
        if style != 'simple':
            lines.append(divider)
    if style == 'simple':
        lines.append(divider)
    return lines


def get_docs_section(docs, keys=['Args:', 'Arguments:'], join_lines=False):
    r"""Get contents of a particular sections in a docstring.

    Args:
        docs (str): Docstring that should be parsed.
        keys (list): Strings that should be used to identify the target section
            in the docstring. Defaults to ['Args:', 'Arguments:'].
        join_lines (bool, optional): If True, the returned lines will be
            stripped and joined into a single string. Defaults to False.

    Returns:
        list: Lines in the section(s) specified by keys.

    """
    if not isinstance(keys, list):
        keys = [keys]
    if docs is None:
        return []
    docs_lines = docs.splitlines()
    # Isolate arguments section based on heading
    in_args = False
    args_lines = []
    for x in docs_lines:
        if in_args:
            if (len(x.strip()) == 0) or (not x.startswith(8 * ' ')):
                # Blank line or no indent indicates new section
                in_args = False
                break
            else:
                args_lines.append(x)
        elif any([x.startswith('    %s' % k) for k in keys]):
            in_args = True
    # Join lines
    if join_lines:
        return ' '.join([x.strip() for x in args_lines])
    return args_lines
    

def docs2args(docs, keys=['Args:', 'Arguments:']):
    r"""Get a dictionary of arguments and argument descriptions from a docstring.

    Args:
        docs (str): Docstring that should be parsed.
        keys (list): Strings that should be used to identify the target section
            in the docstring. Defaults to ['Args:', 'Arguments:'].

    Returns:
        dict: Dictionary of arguments/description pairs.

    """
    args_lines = get_docs_section(docs, keys=keys)
    # Parse argument lines
    out = {}
    curr_arg = None
    for x in args_lines:
        if x.startswith(12 * ' '):
            out[curr_arg]['description'] += ' ' + x.strip()
        else:
            re_arg = r'        ([\S]+)[\s]+\(([^\)]+)\):[\s]+([\S\s]+)'
            x_match = re.match(re_arg, x)
            if x_match is None:
                break
            curr_arg = x_match.group(1)
            out[curr_arg] = {'type': x_match.group(2),
                             'description': x_match.group(3)}
    return out



from yggdrasil.communication.FileComm import FileComm


class AsciiTableComm(FileComm):
    r"""Class for handling I/O from/to a file on disk."""
    
    _filetype = 'table'
    _schema_subtype_description = ('The file is an ASCII table that will be '
                                   'read/written one row at a time. If '
                                   '``as_array`` is ``True``, the table will '
                                   'be read/written all at once.')
    _default_serializer = 'table'



#!/usr/bin/python
from yggdrasil.interface.YggInterface import YggInput, YggOutput


if __name__ == '__main__':

    # Get input and output channels matching yaml
    in1 = YggInput('input1_python')
    in2 = YggInput('static_python')
    out1 = YggOutput('output_python')
    print('SaM(P): Set up I/O channels')

    # Get input from input1 channel
    ret, adata = in1.recv()
    if not ret:
        raise RuntimeError('SaM(P): ERROR RECV from input1')
    a = int(adata)
    print('SaM(P): Received %d from input1' % a)

    # Get input from static channel
    ret, bdata = in2.recv()
    if not ret:
        raise RuntimeError('SaM(P): ERROR RECV from static')
    b = int(bdata)
    print('SaM(P): Received %d from static' % b)

    # Compute sum and send message to output channel
    sum = a + b
    outdata = '%d' % sum
    ret = out1.send(outdata)
    if not ret:
        raise RuntimeError('SaM(P): ERROR SEND to output')
    print('SaM(P): Sent to output')



from __future__ import print_function
import sys
import numpy as np
from yggdrasil.interface.YggInterface import (
    YggRpcClient, YggInput, YggOutput)


def fibClient(args):
    
    iterations = int(args[0])
    print('Hello from Python rpcFibCli: iterations = %d ' % iterations)

    # Set up connections matching yaml
    # RPC client-side connection will be $(server_name)_$(client_name)
    ymlfile = YggInput("yaml_in")
    rpc = YggRpcClient("rpcFibSrv_rpcFibCli", "%d", "%d %d")
    log = YggOutput("output_log")

    # Read entire contents of yaml
    ret, ycontent = ymlfile.recv()
    if not ret:
        raise RuntimeError('rpcFibCli(P): RECV ERROR')
    print('rpcFibCli: yaml has %d lines' % len(ycontent.split(b'\n')))

    for i in range(1, iterations + 1):
        
        # Call the server and receive response
        print('rpcFibCli(P): fib(->%-2d) ::: ' % i, end='')
        ret, fib = rpc.rpcCall(np.int32(i))
        if not ret:
            raise RuntimeError('rpcFibCli(P): RPC CALL ERROR')

        # Log result by sending it to the log connection
        s = 'fib(%2d<-) = %-2d<-\n' % tuple(fib)
        print(s, end='')
        ret = log.send(s)
        if not ret:
            raise RuntimeError('rpcFibCli(P): SEND ERROR')

    print('Goodbye from Python rpcFibCli')

    
if __name__ == '__main__':
    fibClient(sys.argv[1:])



import numpy as np
from yggdrasil import serialize, backwards, platform
from yggdrasil.tests import assert_raises, assert_equal


unsupported_nptype = ['bool_']
map_nptype2cformat = [
    (['float_', 'float16', 'float32', 'float64'], '%g'),
    (['complex_', 'complex64', 'complex128'], '%g%+gj'),
    # (['int8', 'short', 'intc', 'int_', 'longlong'], '%d'),
    # (['uint8', 'ushort', 'uintc', 'uint64', 'ulonglong'], '%u'),
    ('int8', '%hhd'), ('short', '%hd'), ('intc', '%d'),
    ('uint8', '%hhu'), ('ushort', '%hu'), ('uintc', '%u'),
    ('S', '%s'), ('S5', '%5s'), ('U', '%s'), ('U5', '%20s')]
if platform._is_win:  # pragma: windows
    map_nptype2cformat.append(('int64', '%l64d'))
    map_nptype2cformat.append(('uint64', '%l64u'))
else:
    map_nptype2cformat.append(('int64', '%ld'))
    map_nptype2cformat.append(('uint64', '%lu'))
# Conditional on if default int 32bit or 64bit
# This is for when default int is 32bit
if np.dtype('int_') != np.dtype('intc'):
    map_nptype2cformat.append(('int_', '%ld'))
else:
    map_nptype2cformat.append(('int_', '%d'))  # pragma: windows
if np.dtype('int_') != np.dtype('longlong'):
    if platform._is_win:  # pragma: windows
        map_nptype2cformat.append(('longlong', '%l64d'))
        map_nptype2cformat.append(('ulonglong', '%l64u'))
    else:  # pragma: debug
        map_nptype2cformat.append(('longlong', '%lld'))
        map_nptype2cformat.append(('ulonglong', '%llu'))
map_cformat2pyscanf = [(['%5s', '%s'], '%s'),
                       ('%s', '%s'),
                       # (['%hhd', '%hd', '%d', '%ld', '%lld', '%l64d'], '%d'),
                       # (['%hhu', '%hu', '%u', '%lu', '%llu', '%l64u'], '%u'),
                       ('%g%+gj', '%g%+gj')]

unsupported_cfmt = ['a', 'A', 'p', 'n', '']
map_cformat2nptype = [(['f', 'F', 'e', 'E', 'g', 'G'], 'float64'),
                      # (['f', 'F', 'e', 'E', 'g', 'G'], 'float32'),
                      # (['lf', 'lF', 'le', 'lE', 'lg', 'lG'], 'float64'),
                      (['hhd', 'hhi'], 'int8'),
                      (['hd', 'hi'], 'short'),
                      (['d', 'i'], 'intc'),
                      (['ld', 'li'], 'int_'),
                      (['lld', 'lli', 'l64d'], 'longlong'),
                      (['hhu', 'hho', 'hhx', 'hhX'], 'uint8'),
                      (['hu', 'ho', 'hx', 'hX'], 'ushort'),
                      (['u', 'o', 'x', 'X'], 'uintc'),
                      (['lu', 'lo', 'lx', 'lX'], 'uint64'),
                      (['llu', 'llo', 'llx', 'llX', 'l64u'], 'ulonglong'),
                      (['c', 's'], backwards.np_dtype_str),
                      ('s', backwards.np_dtype_str)]
# if np.dtype('int_') != np.dtype('intc'):
#     map_cformat2nptype.append((['ld', 'li'], 'int_'))
map_cformat2nptype.append(
    (['%{}%+{}j'.format(_, _) for _ in ['f', 'F', 'e', 'E', 'g', 'G']],
     'complex128'))


def test_extract_formats():
    r"""Test extract_formats."""
    test_str = ['%10s\t%5.2f\t%4d\t%g%+gj']
    test_fmt = [['%10s', '%5.2f', '%4d', '%g%+gj']]
    for s, f in zip(test_str, test_fmt):
        assert_equal(serialize.extract_formats(s), f)
        assert_equal(serialize.extract_formats(backwards.as_bytes(s)),
                     [backwards.as_bytes(i) for i in f])


def test_nptype2cformat():
    r"""Test conversion from numpy dtype to C format string."""
    for a, b in map_nptype2cformat:
        if isinstance(a, str):
            a = [a]
        for ia in a:
            assert_equal(serialize.nptype2cformat(ia), b)
    assert_raises(TypeError, serialize.nptype2cformat, 0)
    for a in unsupported_nptype:
        assert_raises(ValueError, serialize.nptype2cformat, a)


def test_nptype2cformat_structured():
    r"""Test conversion from structured numpy dtype to C format string."""
    if platform._is_win:  # pragma: windows
        fmts = ["%5s", "%l64d", "%g", "%g%+gj"]
    else:
        fmts = ["%5s", "%ld", "%g", "%g%+gj"]
    names0 = ['f0', 'f1', 'f2', 'f3']
    names1 = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype0 = np.dtype([(n, f) for n, f in zip(names0, dtypes)])
    dtype1 = np.dtype([(n, f) for n, f in zip(names1, dtypes)])
    alist = [dtype0, dtype1]
    blist = [fmts, fmts]
    for a, b in zip(alist, blist):
        assert_equal(serialize.nptype2cformat(a), b)
        assert_equal(serialize.nptype2cformat(a, asbytes=True),
                     [backwards.as_bytes(ib) for ib in b])


def test_cformat2nptype():
    r"""Test conversion from C format string to numpy dtype."""
    for a, b in map_cformat2nptype:
        if isinstance(a, str):
            a = [a]
        for _ia in a:
            if _ia.startswith(backwards.as_str(serialize._fmt_char)):
                ia = backwards.as_bytes(_ia)
            else:
                ia = serialize._fmt_char + backwards.as_bytes(_ia)
            assert_equal(serialize.cformat2nptype(ia), np.dtype(b))  # .str)
            # assert_equal(serialize.cformat2nptype(ia), np.dtype(b).str)
    assert_raises(TypeError, serialize.cformat2nptype, 0)
    assert_raises(ValueError, serialize.cformat2nptype, b's')
    assert_raises(ValueError, serialize.cformat2nptype, b'%')
    assert_raises(ValueError, serialize.cformat2nptype,
                  '%d\t%f', names=['one'])
    for a in unsupported_nptype:
        assert_raises(ValueError, serialize.cformat2nptype,
                      backwards.as_bytes('%' + a))


def test_cformat2nptype_structured():
    r"""Test conversion from C format string to numpy dtype for structured
    data types."""
    if platform._is_win:  # pragma: debug
        fmts = ["%5s", "%l64d", "%lf", "%g%+gj"]
    else:
        fmts = ["%5s", "%ld", "%lf", "%g%+gj"]
    names0 = ['f0', 'f1', 'f2', 'f3']
    names1 = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype0 = np.dtype([(n, f) for n, f in zip(names0, dtypes)])
    dtype1 = np.dtype([(n, f) for n, f in zip(names1, dtypes)])
    alist = ["\t".join(fmts) + "\n", ''.join(fmts), fmts]
    for a in alist:
        b0 = serialize.cformat2nptype(a)
        assert_equal(b0, dtype0)
        b1 = serialize.cformat2nptype(a, names=names1)
        assert_equal(b1, dtype1)


def test_cformat2pyscanf():
    r"""Test conversion of C format string to version for python scanf."""
    all_a = []
    all_b = []
    for a, b in map_cformat2pyscanf:
        if isinstance(a, str):
            a = [a]
        for _ia in a:
            ia = backwards.as_bytes(_ia)
            ib = backwards.as_bytes(b)
            all_a.append(ia)
            all_b.append(ib)
            assert_equal(serialize.cformat2pyscanf(ia), ib)
    assert_raises(TypeError, serialize.cformat2pyscanf, 0)
    assert_raises(ValueError, serialize.cformat2pyscanf, b's')
    assert_raises(ValueError, serialize.cformat2pyscanf, b'%')
    fmt_a = b'\t'.join(all_a)
    fmt_b = b'\t'.join(all_b)
    assert_equal(serialize.cformat2pyscanf(all_a), all_b)
    assert_equal(serialize.cformat2pyscanf(fmt_a), fmt_b)


def test_format_message():
    r"""Test formatting message from a list or arguments and back."""
    fmt = b'%5s\t%ld\t%lf\t%g%+gj\n'
    dtype = serialize.cformat2nptype(fmt)
    x_arr = np.ones(1, dtype)
    x_tup = [x_arr[n][0] for n in x_arr.dtype.names]
    # x_tup[0] = backwards.as_str(x_tup[0])
    flist = [fmt, "%ld"]
    alist = [tuple(x_tup), 0]
    for a, f in zip(alist, flist):
        msg = serialize.format_message(a, f)
        b = serialize.process_message(msg, f)
        if not isinstance(a, tuple):
            assert_equal(b, (a, ))
        else:
            assert_equal(b, a)
    # Errors
    assert_raises(RuntimeError, serialize.format_message, (0, ), "%d %d")
    assert_raises(TypeError, serialize.process_message, 0, "%d")
    assert_raises(ValueError, serialize.process_message, b'hello', "%d")


def test_combine_flds():
    r"""Test combine_flds."""
    names0 = ['f0', 'f1', 'f2', 'f3']
    names1 = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype0 = np.dtype([(n, f) for n, f in zip(names0, dtypes)])
    dtype1 = np.dtype([(n, f) for n, f in zip(names1, dtypes)])
    nfld = len(names0)
    nele = 5
    arrs = [np.zeros(nele, dtype=dtype0[i]) for i in range(nfld)]
    np.testing.assert_array_equal(serialize.combine_flds(arrs),
                                  np.zeros(nele, dtype0))
    np.testing.assert_array_equal(serialize.combine_flds(arrs, dtype=dtype1),
                                  np.zeros(nele, dtype1))
    assert_raises(ValueError, serialize.combine_flds, arrs[:-1], dtype=dtype0)
    arrs[0] = np.zeros(nele - 1, dtype=dtype0[0])
    assert_raises(ValueError, serialize.combine_flds, arrs)


def test_combine_eles():
    r"""Test combine_eles."""
    names0 = ['f0', 'f1', 'f2', 'f3']
    names1 = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype0 = np.dtype([(n, f) for n, f in zip(names0, dtypes)])
    dtype1 = np.dtype([(n, f) for n, f in zip(names1, dtypes)])
    dtype_short = np.dtype([(n, f) for n, f in zip(names1[:-1], dtypes[:-1])])
    # nfld = len(names0)
    nele = 5
    res0 = np.zeros(nele, dtype0)
    res1 = np.zeros(nele, dtype1)
    res0['f0'][0] = 'hello'
    res1['name'][0] = 'hello'
    arrs = [np.zeros(1, dtype=dtype0) for i in range(nele)]
    arrs[0]['f0'] = b'hello'
    arrs_list = [a.tolist()[0] for a in arrs]
    arrs_void = [res1[i] for i in range(nele)]
    arrs_mixd = [a.tolist() for a in res1]
    arrs_mixd[-1] = arrs_void[-1]
    if platform._is_win:  # pragma: windows
        if backwards.PY2:  # pragma: Python 2
            res0_list = res0
            res1_list = res1
        else:  # pragma: Python 3
            dtype0_w = np.dtype({'names': names0, 'formats': ['S5', 'i4', 'f8', 'c16']})
            dtype1_w = np.dtype({'names': names1, 'formats': ['S5', 'i4', 'f8', 'c16']})
            res0_list = res0.astype(dtype0_w)
            res1_list = res1.astype(dtype1_w)
    else:
        res0_list = res0
        res1_list = res1
    np.testing.assert_array_equal(serialize.combine_eles([res1[n][0] for n in names1]),
                                  res0[0])
    np.testing.assert_array_equal(serialize.combine_eles(arrs), res0)
    np.testing.assert_array_equal(serialize.combine_eles(arrs_list), res0_list)
    np.testing.assert_array_equal(serialize.combine_eles(arrs_void), res1)
    np.testing.assert_array_equal(serialize.combine_eles(arrs_mixd), res1_list)
    np.testing.assert_array_equal(serialize.combine_eles(arrs, dtype=dtype1),
                                  res1)
    np.testing.assert_array_equal(serialize.combine_eles(arrs_list, dtype=dtype1),
                                  res1)
    arrs_list[0] = arrs_list[0][:-1]
    assert_raises(ValueError, serialize.combine_eles, arrs_list, dtype=dtype1)
    assert_raises(ValueError, serialize.combine_eles, arrs_list)
    arrs_list[0] = None
    assert_raises(TypeError, serialize.combine_eles, arrs_list)
    assert_raises(ValueError, serialize.combine_eles, arrs_void,
                  dtype=dtype_short)


def test_consolidate_array():
    r"""Test consolidation of array information in different forms."""
    names0 = ['f0', 'f1', 'f2', 'f3']
    # names1 = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype0 = np.dtype([(n, f) for n, f in zip(names0, dtypes)])
    if platform._is_win:  # pragma: windows
        if backwards.PY2:  # pragma: Python 2
            dtype0_list = dtype0
        else:  # pragma: Python 3
            dtype0_list = np.dtype({'names': names0,
                                    'formats': ['S5', 'i4', 'f8', 'c16']})
    else:
        dtype0_list = dtype0
    # dtype1 = np.dtype([(n, f) for n, f in zip(names1, dtypes)])
    dtype2 = np.dtype([('f0', 'i8'), ('f1', 'f8')])
    dtype3 = np.dtype([('f0', 'i4'), ('f1', 'f4')])
    shape = (5, 5)
    # Create list of input variables
    dlist = []
    x0list = []
    x1list = []
    for dtype in [np.dtype('float'), dtype0]:
        x = np.zeros(shape, dtype=dtype)
        if dtype == dtype0:
            x['f0'][:] = 'hello'
        x_flat = x.flatten()
        dlist.append(dtype)
        x0list.append(x)
        x1list.append(x)
        if len(dtype) > 0:
            dlist += [dtype, dtype]
            if dtype == dtype0:
                dtype_list = dtype0_list
            else:  # pragma: debug
                dtype_list = dtype
            dlist.append(dtype_list)
            x0list += [x, x_flat, x_flat.astype(dtype_list)]
            # Tests with lists of arrays rows or columns
            x1list.append([x[n] for n in dtype.names])
            x1list.append([x_flat[i] for i in range(len(x_flat))])
            x1list.append([x_flat[i].tolist() for i in range(len(x_flat))])
    # Tests with single array mapped onto structured array
    nrow = 4
    x0 = np.zeros(nrow, dtype2)
    x1 = np.zeros((nrow, len(dtype2)))
    dlist += [None]
    x0list += [x0]
    x1list.append(x0)
    x2 = serialize.consolidate_array(x1, dtype=dtype2)
    np.testing.assert_array_equal(x0, x2)
    np.testing.assert_array_equal(x0, serialize.consolidate_array(
        np.zeros(nrow, dtype3), dtype=dtype2))
    # Test with list of arrays
    arr_void = np.ones(5, dtype0)
    x0list.append(arr_void)
    x1list.append([a for a in arr_void])
    dlist.append(dtype0)
    # Loop over different test inputs
    i = 0
    for x0, x1, dtype in zip(x0list, x1list, dlist):
        i += 1
        # Dtype provided
        x2 = serialize.consolidate_array(x1, dtype=dtype)
        np.testing.assert_array_equal(x0, x2)
        # Dtype not provided
        x2 = serialize.consolidate_array(x1)
        np.testing.assert_array_equal(x0, x2)
    # Error on incorrect data format
    assert_raises(ValueError, serialize.consolidate_array,
                  np.zeros((4, 1)), dtype=dtype2)
    # Error on incorrect type
    assert_raises(TypeError, serialize.consolidate_array, None)
    # Error on dtypes with differing numbers of fields
    assert_raises(ValueError, serialize.consolidate_array,
                  np.zeros(3, dtype0), dtype=dtype3)
                     

def test_format2table():
    r"""Test getting table information from a format string."""
    out = {'delimiter': b'\t',
           'newline': b'\n',
           'comment': b'# ',
           'fmts': ["%5s", "%ld", "%lf", "%g%+gj"]}
    out['fmts'] = [backwards.as_bytes(f) for f in out['fmts']]
    sfmt = out['fmts'][0]
    sout = dict(**out)
    sout['fmts'] = [sfmt]
    del sout['newline'], sout['comment']
    fmt = b'# %5s\t%ld\t%lf\t%g%+gj\n'
    fmt2 = b'# %5s\t\t%ld\t%lf\t%g%+gj\n'
    assert_equal(dict(fmts=[]), serialize.format2table('hello'))
    assert_equal(sout, serialize.format2table(sfmt))
    assert_equal(fmt, serialize.table2format(**out))
    assert_equal(out, serialize.format2table(fmt))
    assert_equal(fmt, serialize.table2format(fmts=out['fmts']))
    assert_equal(out, serialize.format2table(fmt2))
    assert_raises(RuntimeError, serialize.format2table, "%5s,%ld\t%g\n")


def test_array_to_table():
    r"""Test conversion of arrays to ASCII table and back."""
    flist = ['# %5s\t%ld\t%lf\t%g%+gj\n']
    for use_astropy in [False, True]:
        for f in flist:
            dtype = serialize.cformat2nptype(f)
            arr0 = np.ones(5, dtype)
            arr0['f0'][0] = b'hello'
            tab = serialize.array_to_table(arr0, f, use_astropy=use_astropy)
            arr1 = serialize.table_to_array(tab, f, use_astropy=use_astropy)
            np.testing.assert_array_equal(arr1, arr0)


def test_array_to_bytes():
    r"""Test conversion of arrays to bytes and back."""
    names0 = ['f0', 'f1', 'f2', 'f3']
    # names1 = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype0 = np.dtype([(n, f) for n, f in zip(names0, dtypes)])
    # dtype1 = np.dtype([(n, f) for n, f in zip(names1, dtypes)])
    shape = (5, 5)
    for order in ['C', 'F']:
        for dtype in [np.dtype('float'), dtype0]:
            x0 = np.zeros(shape, dtype=dtype)
            if dtype == dtype0:
                x0['f0'][:] = 'hello'
            # An array of the same type, dtype provided
            b = serialize.array_to_bytes(x0, dtype=dtype, order=order)
            x1 = serialize.bytes_to_array(b, dtype, order=order, shape=shape)
            np.testing.assert_array_equal(x1, x0)
            # An array of the same type, dtype not provided
            b = serialize.array_to_bytes(x0, order=order)
            x1 = serialize.bytes_to_array(b, dtype, order=order, shape=shape)
            np.testing.assert_array_equal(x1, x0)
        # Tests with single array mapped onto structured array
        dtype2 = np.dtype([('f0', 'i8'), ('f1', 'f8')])
        nrow = 4
        ncol = len(dtype2)
        b0 = serialize.array_to_bytes(np.zeros(nrow, dtype2), order=order)
        b1 = serialize.array_to_bytes(np.zeros((nrow, ncol)), dtype=dtype2,
                                      order=order)
        assert_equal(b1, b0)
        # Error on incomplete serial array
        assert_raises(RuntimeError, serialize.bytes_to_array,
                      b0[:-1], dtype2, order=order)


def test_format_header():
    r"""Test formatting header."""
    kws_all = dict(
        field_names=['name', 'number', 'value', 'complex'],
        field_units=['n/a', 'g', 'cm', 'n/a'])
    res_all = dict(
        names="# name\tnumber\tvalue\tcomplex\n",
        units="# n/a\tg\tcm\tn/a\n")
    if platform._is_win:  # pragma: windows
        kws_all['format_str'] = "%5s\t%l64d\t%g\t%g%+gj\n"
        res_all['format'] = "# " + kws_all['format_str']
    else:
        kws_all['format_str'] = "%5s\t%ld\t%g\t%g%+gj\n"
        res_all['format'] = "# " + kws_all['format_str']
    kws_all['dtype'] = serialize.cformat2nptype(kws_all['format_str'],
                                                names=kws_all['field_names'])
    for x in [kws_all, res_all]:
        for k, v in x.items():
            if isinstance(v, str):
                x[k] = backwards.as_bytes(v)
            elif isinstance(v, list):
                x[k] = [backwards.as_bytes(iv) for iv in v]
    test_list = [(['format_str', 'field_names', 'field_units'],
                  ['names', 'units', 'format']),
                 (['field_names', 'field_units', 'dtype'],
                  ['names', 'units', 'format']),
                 (['field_units', 'dtype'],
                  ['names', 'units', 'format']),
                 (['field_names'], ['names']),
                 (['field_units'], ['units'])]
    for kws_keys, res_keys in test_list:
        kws = {k: kws_all[k] for k in kws_keys}
        res = b''.join([res_all[k] for k in res_keys])
        assert_equal(serialize.format_header(**kws), res)
    assert_raises(ValueError, serialize.format_header)


def test_parse_header():
    r"""Test parsing header."""
    header = ["# name\tnumber\tvalue\tcomplex\n",
              "# n/a\tg\tcm\tn/a\n",
              "# %5s\t%ld\t%lf\t%g%+gj\n"]
    res = dict(delimiter='\t', comment='# ', newline='\n',
               format_str="%5s\t%ld\t%lf\t%g%+gj\n",
               fmts=['%5s', '%ld', '%lf', '%g%+gj'],
               field_names=['name', 'number', 'value', 'complex'],
               field_units=['n/a', 'g', 'cm', 'n/a'])
    for i in range(len(header)):
        header[i] = backwards.as_bytes(header[i])
    for k, v in res.items():
        if isinstance(v, str):
            res[k] = backwards.as_bytes(v)
        elif isinstance(v, list):
            res[k] = [backwards.as_bytes(s) for s in v]
    assert_equal(serialize.parse_header(header), res)
    assert_equal(serialize.parse_header(header[::-1]), res)
    _empty = b''
    assert_equal(serialize.parse_header(_empty.join(header)), res)
    # Test without formats
    header2 = header[:2]
    res2 = dict(**res)
    del res2['format_str']
    res2['fmts'] = []
    assert_equal(serialize.parse_header(header2), res2)
    # Test with explicit line numbers
    assert_equal(serialize.parse_header(header, lineno_names=0, lineno_units=1),
                 res)
    # Test errors
    header3 = [header[0], header[0]]
    assert_raises(RuntimeError, serialize.parse_header, header3)
    header4 = [header[1], header[1]]
    assert_raises(RuntimeError, serialize.parse_header, header4)


def test_dict2list():
    r"""Test conversion of a dictionary to a list and back."""
    assert_raises(TypeError, serialize.dict2list, None)
    assert_raises(TypeError, serialize.list2dict, None)


def test_numpy2pandas():
    r"""Test conversion of a numpy array to a pandas data frame and back."""
    assert_raises(TypeError, serialize.numpy2pandas, None)
    assert_raises(TypeError, serialize.pandas2numpy, None)
    nele = 5
    names = ["name", "number", "value", "complex"]
    dtypes = ['S5', 'i8', 'f8', 'c16']
    dtype = np.dtype([(n, f) for n, f in zip(names, dtypes)])
    arr_mix = np.zeros(nele, dtype)
    arr_mix['name'][0] = 'hello'
    arr_obj = np.array([list(), 'hello', 5], dtype='O')
    test_arrs = [arr_mix,
                 np.zeros(nele, 'float'),
                 arr_mix['name'],
                 arr_obj,
                 np.array([])]
    for ans in test_arrs:
        frame = serialize.numpy2pandas(ans)
        res = serialize.pandas2numpy(frame)
        np.testing.assert_array_equal(ans, res)


def test_numpy2list():
    r"""Test conversion of a numpy array to a list and back."""
    assert_raises(TypeError, serialize.numpy2list, None)
    assert_raises(TypeError, serialize.list2numpy, None)
    nele = 5
    names = ["complex", "name", "number", "value"]
    dtypes = ['c16', 'S5', 'i8', 'f8']
    dtype = np.dtype([(n, f) for n, f in zip(names, dtypes)])
    arr_mix = np.zeros(nele, dtype)
    arr_mix['name'][0] = 'hello'
    test_arrs = [arr_mix]
    for ans in test_arrs:
        d = serialize.numpy2list(ans)
        # Name provided
        res = serialize.list2numpy(d, names=names)
        np.testing.assert_array_equal(ans, res)


def test_numpy2dict():
    r"""Test conversion of a numpy array to a dictionary and back."""
    assert_raises(TypeError, serialize.numpy2dict, None)
    assert_raises(TypeError, serialize.dict2numpy, None)
    nele = 5
    names = ["complex", "name", "number", "value"]
    dtypes = ['c16', 'S5', 'i8', 'f8']
    dtype = np.dtype([(n, f) for n, f in zip(names, dtypes)])
    arr_mix = np.zeros(nele, dtype)
    arr_mix['name'][0] = 'hello'
    test_arrs = [arr_mix]
    for ans in test_arrs:
        d = serialize.numpy2dict(ans)
        # Sorted
        res = serialize.dict2numpy(d)
        np.testing.assert_array_equal(ans, res)
        # Provided
        res = serialize.dict2numpy(d, order=names)
        np.testing.assert_array_equal(ans, res)


def test_pandas2dict():
    r"""Test conversion of a Pandas data frame to a dictionary and back."""
    assert_raises(TypeError, serialize.dict2pandas, None)
    assert_raises(TypeError, serialize.pandas2dict, None)
    nele = 5
    names = ["complex", "name", "number", "value"]
    dtypes = ['c16', 'S5', 'i8', 'f8']
    dtype = np.dtype([(n, f) for n, f in zip(names, dtypes)])
    arr_mix = np.zeros(nele, dtype)
    arr_mix['name'][0] = 'hello'
    arr_obj = np.array([list(), 'hello', 5], dtype='O')
    test_arrs = [arr_mix,
                 np.zeros(nele, 'float'),
                 arr_mix['name'],
                 arr_obj,
                 np.array([])]
    for ans in test_arrs:
        frame = serialize.numpy2pandas(ans)
        # Sorted
        d = serialize.pandas2dict(frame)
        res = serialize.dict2pandas(d)
        np.testing.assert_array_equal(res, frame)
        # Provided
        d = serialize.pandas2dict(frame)
        res = serialize.dict2pandas(d, order=ans.dtype.names)
        np.testing.assert_array_equal(res, frame)
